"""
Periodic Linkerless Honeycomb
=============================

"""


import numpy as np

from stk._internal.optimizers.null import NullOptimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.topology_graphs.edge import Edge

from .cof import Cof
from .vertices import NonLinearVertex


class PeriodicLinkerlessHoneycomb(Cof):
    """
    Represents a periodic honeycomb COF topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.PeriodicLinkerlessHoneycomb(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='BrCC(CBr)CBr',
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                lattice_size=(3, 3, 1),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
                if all(p == 0 for p in bond.get_periodicity())
            ),
        )

    ``Collapser(scale_steps=False)`` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.PeriodicLinkerlessHoneycomb(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='BrCC(CBr)CBr',
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                lattice_size=(3, 3, 1),
                optimizer=stk.Collapser(scale_steps=False),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
                if all(p == 0 for p in bond.get_periodicity())
            ),
        )

    Building blocks with three functional groups are required
    for this topology graph.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cof-topology-graph-examples`:
    *Multi-Building Block COF Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 1

    Note that optimizers may not optimize the :class:`.PeriodicInfo`.
    The documentation of the optimizer will state if it does.

    See :class:`.Cof` for more details and examples.

    """

    def __init__(
        self,
        building_blocks,
        lattice_size,
        vertex_alignments=None,
        reaction_factory=GenericReactionFactory(),
        num_processes=1,
        optimizer=NullOptimizer(),
    ):
        """
        Initialize a :class:`.PeriodicLinkerlessHoneycomb` instance.

        Parameters
        ----------
        building_blocks : :class:`tuple` or :class:`dict`
            Can be a :class:`tuple` of :class:`.BuildingBlock`
            instances, which should be placed on the topology graph.

            Can also be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on. A :class:`dict` is
            required when there are multiple building blocks with the
            same number of functional groups, because in this case
            the desired placement is ambiguous.

        lattice_size : :class:`tuple` of :class:`int`
            The size of the lattice in the x, y and z directions.

        vertex_alignments : :class:`dict`, optional
            A mapping from the id of a :class:`.Vertex`
            to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is referred to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        reaction_factory : :class:`.ReactionFactory`, optional
            The reaction factory to use for creating bonds between
            building blocks.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        optimizer : :class:`.Optimizer`, optional
            Used to optimize the structure of the constructed
            molecule.

        Raises
        ------
        :class:`AssertionError`
            If the any building block does not have a
            valid number of functional groups.

        :class:`ValueError`
            If the there are multiple building blocks with the
            same number of functional_groups in `building_blocks`,
            and they are not explicitly assigned to vertices. The
            desired placement of building blocks is ambiguous in
            this case.

        :class:`~.cof.UnoccupiedVertexError`
            If a vertex of the COF topology graph does not have a
            building block placed on it.

        :class:`~.cof.OverlyOccupiedVertexError`
            If a vertex of the COF topology graph has more than one
            building block placed on it.

        """

        super().__init__(
            building_blocks=building_blocks,
            lattice_size=lattice_size,
            periodic=True,
            vertex_alignments=vertex_alignments,
            reaction_factory=reaction_factory,
            num_processes=num_processes,
            optimizer=optimizer,
        )

    _lattice_constants = _a, _b, _c = (
        np.array([1.0, 0.0, 0.0]),
        np.array([0.5, 0.866, 0.0]),
        np.array([0.0, 0.0, 5 / 1.7321]),
    )

    _vertex_prototypes = (
        NonLinearVertex(0, (1 / 3) * _a + (1 / 3) * _b + (1 / 2) * _c),
        NonLinearVertex(1, (2 / 3) * _a + (2 / 3) * _b + (1 / 2) * _c),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(
            id=1,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            periodicity=(-1, 0, 0),
        ),
        Edge(
            id=2,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            periodicity=(0, -1, 0),
        ),
    )



import numpy as np


def test_get_position_matrix(case_data):
    """
    Test :meth:`.Molecule.get_position_matrix`.

    Parameters
    ----------
    case_data : :class:`.CaseData
        A test case. Holds the molecule to test and the correct
        position matrix.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert np.allclose(
        a=case_data.position_matrix,
        b=case_data.molecule.get_position_matrix(),
        atol=1e-6,
    )



import typing

from stk._internal.ea.molecule_record import MoleculeRecord

from .selector import ExcludedBatches, IncludedBatches, Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class FilterBatches(Selector[T]):
    """
    Allows a :class:`.Selector` to select only some batches.

    Examples:

        *Using a Selection Algorithm on a Subset of Batches*

        You only want the :class:`.Best` 10 batches to participate in
        :class:`.Roulette`

        .. testcode:: using-a-selection-algorithm-on-a-subset-of-batches

            import stk

            # Select the 10 best batches first, and then run roulette on
            # those.
            selector = stk.FilterBatches(
                filter=stk.Best(10),
                selector=stk.Roulette(7),
            )

            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(100)
            }

            for batch in selector.select(population):
                # Do stuff with batch. It is one of the 10 best batches and
                # was selected using roulette selection.
                pass
    """

    def __init__(self, filter: Selector[T], selector: Selector[T]) -> None:
        """
        Parameters:
            filter:
                Selects batches which can be yielded by `selector`.

            selector:
                Selects batches, but only if they were also selected by
                `filter`.
        """
        self._filter = filter
        self._selector = selector

    def select(
        self,
        population: dict[T, float],
        included_batches: "IncludedBatches" = None,
        excluded_batches: "ExcludedBatches" = None,
    ):
        allowed = {
            batch.get_identity_key()
            for batch in self._filter.select(
                population=population,
                included_batches=included_batches,
                excluded_batches=excluded_batches,
            )
        }
        yield from self._selector.select(
            population=population,
            included_batches=allowed,
            excluded_batches=excluded_batches,
        )



import typing
from collections.abc import Callable
from functools import partial
from typing import Any

import numpy as np

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class ShiftUp(FitnessNormalizer[T]):
    """
    Shifts negative fitness values to be positive.

    Assume you have a vector-valued fitness value, where each number
    represents a different property of the molecule, for example,
    ``[1, -10, 1]``

    One way to convert the vector-valued fitness value into a
    scalar fitness value is by summing the elements, and the result in
    this case would be ``-8``. Clearly this doesn't work, because the
    resulting fitness value is not a positive number. To fix this,
    the ``-10`` should be shifted to a positive value.

    :class:`.ShiftUp` finds the minimum value of each element in the
    vector-valued fitness value across the entire population, and for
    elements where this minimum value is less than ``0``, shifts up
    the element value for every molecule in the population, so that the
    minimum value in the entire population is ``1``.

    For example, take a population with the vector-valued fitness
    values

    .. code-block:: python

            [1, -5, 5]
            [3, -10, 2]
            [2, 20, 1]

    After normalization the fitness values will be.

    .. code-block:: python

            [1, 6, 5]
            [3, 1, 2]
            [2, 31, 1]

    :class:`.ShiftUp` also works when the fitness value is a
    single value.

    Examples:

        *Ensuring Positive Fitness Values*

        Here you final fitness value is calculated by taking a
        :class:`.Sum` of the different components of the fitness value.
        To ensure that the final sum is positive, each component must
        also be positive.

        .. testcode:: ensuring-positive-fitness-values

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: (1, -2, 3),
                record2: (4, 5, -6),
            }

            # Create the normalizer.
            shifter = stk.ShiftUp()

            normalized_fitness_values = shifter.normalize(fitness_values)
            assert np.all(np.equal(
                normalized_fitness_values[record1],
                (1, 1, 10),
            ))
            assert np.all(np.equal(
                normalized_fitness_values[record2],
                (4, 8, 1),
            ))

        *Selectively Normalizing Fitness Values*

        Sometimes, you only want to normalize some members of a population,
        for example if some do not have an assigned fitness value,
        because the fitness calculation failed for whatever reason.
        You can use the `filter` parameter to exclude records from the
        normalization

        .. testcode:: selectively-normalizing-fitness-values

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: (1, -2, 3),
                record2: None,
            }
            normalizer = stk.ShiftUp(
                # Only normalize values which are not None.
                filter=lambda fitness_values, record:
                    fitness_values[record] is not None,
            )
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert np.all(np.equal(
                normalized_fitness_values[record1],
                (1, 1, 3),
            ))
            assert normalized_fitness_values[record2] is None
    """

    def __init__(
        self,
        filter: Callable[
            [dict[T, Any], T], bool
        ] = lambda fitness_values, record: True,
    ) -> None:
        """
        Parameters:
            filter:
                A function which returns ``True`` or ``False``. Only
                molecules which return ``True`` will have fitness values
                normalized. By default, all molecules will have fitness
                values normalized.
                The instance passed to the `fitness_values` argument of
                :meth:`.normalize` is passed as the first argument, while
                the second argument will be passed every
                :class:`.MoleculeRecord` in it, one at a time.
        """
        self._filter = filter

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        filtered = filter(
            partial(self._filter, fitness_values),
            fitness_values,
        )
        # Get all the fitness arrays in a matrix.
        fmat = np.array([fitness_values[record] for record in filtered])

        # Get the minimum value of each element across the population.
        # keepdims ensures that np.min returns a 1-D array, because
        # it will be True if fitness values are scalar and False if
        # they are array-valued.
        mins = np.min(fmat, axis=0, keepdims=len(fmat.shape) == 1)

        # Convert all elements in mins which are not to be shifted to 0
        # and make the shift equal to the minimum value + 1.
        shift = np.zeros(len(mins))
        for i, min_ in enumerate(mins):
            if min_ <= 0:
                shift[i] = 1 - min_

        return {
            record: fitness_value + shift
            if self._filter(fitness_values, record)
            else fitness_value
            for record, fitness_value in fitness_values.items()
        }



import numpy as np
import pytest
import stk
from pytest_lazyfixture import lazy_fixture

from ...case_data import CaseData


@pytest.fixture(
    params=(
        lazy_fixture("center"),
        lazy_fixture("head"),
        lazy_fixture("tail"),
        lazy_fixture("unaligning"),
    ),
)
def case_data(request):
    return request.param


@pytest.fixture
def center(id, position, flip):
    return CaseData(
        vertex=stk.polymer.LinearVertex(id, position, flip),
        id=id,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture
def head(id, position, flip):
    return CaseData(
        vertex=stk.polymer.HeadVertex(id, position, flip),
        id=id,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture
def tail(id, position, flip):
    return CaseData(
        vertex=stk.polymer.TailVertex(id, position, flip),
        id=id,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture
def unaligning(id, position, flip):
    return CaseData(
        vertex=stk.polymer.UnaligningVertex(
            id=id,
            position=position,
            flip=flip,
        ),
        id=id,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture(params=(0,))
def id(request):
    return request.param


@pytest.fixture(params=(True, False))
def flip(request):
    return request.param



def test_clone(bond):
    """
    Test :meth:`.Bond.clone`.

    Parameters
    ----------
    bond : :class:`.Bond`
        The bond to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = bond.clone()

    assert clone.get_atom1() is bond.get_atom1()
    assert clone.get_atom2() is bond.get_atom2()
    assert clone.get_periodicity() is bond.get_periodicity()
    assert clone.get_order() is bond.get_order()



import pytest
import stk

from ....case_data import CaseData
from ...building_blocks import get_pd_atom


def _get_copper_atom() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="[Cu+2]",
        functional_groups=(
            stk.SingleAtom(stk.Cu(0, charge=2)) for i in range(4)
        ),
        position_matrix=([0, 0, 0],),
    )


def _get_bi_1() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="O=C(O)c1ccc(Br)cc1",
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#8]~[#1]",
                bonders=(1,),
                deleters=(2,),
            ),
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#8X1]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )


def _get_bi_2() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="Nc1ccc(C(=O)O)cc1",
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#8]~[#1]",
                bonders=(1,),
                deleters=(2,),
            ),
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#8X1]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.Paddlewheel(
                    metals={
                        _get_copper_atom(): (0,),
                        get_pd_atom(): (1,),
                    },
                    ligands={
                        _get_bi_1(): (2, 3),
                        _get_bi_2(): (0, 1),
                    },
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C([H])C(C2=O->[Pd+2]34<-O=C(C5=C([H])C([H])=C("
                "Br)C([H])=C5[H])O->[Cu+2](<-O2)(<-OC(C2=C([H])C([H])"
                "=C(N([H])[H])C([H])=C2[H])=O->3)<-OC(C2=C([H])C([H])"
                "=C(N([H])[H])C([H])=C2[H])=O->4)=C([H])C([H])=C1Br"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.Paddlewheel(
                    metals=_get_copper_atom(),
                    ligands=_get_bi_1(),
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C([H])C(C2=O->[Cu+2]34<-O=C(C5=C([H])C([H])=C("
                "Br)C([H])=C5[H])O->[Cu+2](<-O2)(<-OC(C2=C([H])C([H])"
                "=C(Br)C([H])=C2[H])=O->3)<-OC(C2=C([H])C([H])=C(Br)C"
                "([H])=C2[H])=O->4)=C([H])C([H])=C1Br"
            ),
            name=name,
        ),
    ),
)
def metal_complex_paddlewheel(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
COF Vertices
============

"""

import numpy as np
from scipy.spatial.distance import euclidean

from stk._internal.topology_graphs.vertex import Vertex
from stk._internal.utilities.utilities import get_acute_vector

from ..utilities import _EdgeSorter, _FunctionalGroupSorter


class _CofVertex(Vertex):
    """
    A :class:`._CofVertex` .

    """

    def __init__(self, id, position, aligner_edge=0, cell=(0, 0, 0)):
        """
        Initialize a :class:`._CofVertex`.

        Parameters
        ----------
        id : :class:`int`
            The id of the vertex.

        position : :class:`tuple` of :class:`float`
            The position of the vertex.

        aligner_edge : :class:`int`, optional
            The edge which is used to align the :class:`.BuildingBlock`
            placed on the vertex. The first :class:`.FunctionalGroup`
            is rotated such that it lies exactly on this
            :class:`.Edge`. Must be between ``0`` and the number of
            edges the vertex is connected to.

        cell : :class:`tuple` of :class:`int`, optional
            The cell of the lattice in which the vertex is found.

        """

        super().__init__(id, position)
        self._aligner_edge = aligner_edge
        self._cell = np.array(cell)

    def get_aligner_edge(self):
        """
        Return the aligner edge of the vertex.

        Returns
        -------
        :class:`int`
            The aligner edge.

        """

        return self._aligner_edge

    def get_cell(self):
        return np.array(self._cell)

    @classmethod
    def init_at_center(
        cls,
        id,
        vertices,
        aligner_edge=0,
        cell=(0, 0, 0),
    ):
        """
        Initialize a :class:`._CofVertex` in the middle of `vertices`.

        Parameters
        ----------
        id : :class:`int`
            The id of the initialized vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            The vertices at whose center this one needs to be.

        aligner_edge : :class:`int`, optional
            The edge which is used to align the :class:`.BuildingBlock`
            placed on the vertex. The first :class:`.FunctionalGroup`
            is rotated such that it lies exactly on this
            :class:`.Edge`. Must be between ``0`` and the number of
            edges the vertex is connected to.

        cell : :class:`tuple` of :class:`int`, optional
            The cell of the lattice in which the vertex is found.

        Returns
        -------
        :class:`._CofVertex`
            The new vertex.

        """

        return cls(
            id=id,
            position=(
                sum(vertex.get_position() for vertex in vertices)
                / len(vertices)
            ),
            aligner_edge=aligner_edge,
            cell=cell,
        )

    @classmethod
    def init_at_shifted_center(
        cls,
        id,
        vertices,
        cell_shifts,
        lattice_constants,
        aligner_edge=0,
        cell=(0, 0, 0),
    ):
        """
        Initialize a :class:`._CofVertex` at the center of `vertices`.

        The `vertices` are shifted according to the lattice constants
        and cell shifts.

        Parameters
        ----------
        id : :class:`int`
            The id of the initialized vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            The vertices at whose center this one needs to be.

        cell_shifts : :class:`tuple` of :class:`int`
            The number of cells shifted in the x, y and z directions.

        lattice_constants : :class:`tuple` of :class:`numpy.ndarray`
            The a, b and c lattice constants.

        aligner_edge : :class:`int`, optional
            The edge which is used to align the :class:`.BuildingBlock`
            placed on the vertex. The first :class:`.FunctionalGroup`
            is rotated such that it lies exactly on this
            :class:`.Edge`. Must be between ``0`` and the number of
            edges the vertex is connected to.

        cell : :class:`tuple` of :class:`int`, optional
            The cell of the lattice in which the vertex is found.

        Returns
        -------
        :class:`._CofVertex`
            The new vertex.

        """

        positions = []
        for vertex, cell_shift in zip(vertices, cell_shifts):
            shift = sum(
                dim_shift * constant
                for dim_shift, constant in zip(cell_shift, lattice_constants)
            )
            positions.append(vertex.get_position() + shift)

        position = np.divide(
            np.sum(positions, axis=0),
            len(positions),
        )
        return cls(id, position, aligner_edge, cell)

    def clone(self):
        clone = super().clone()
        clone._aligner_edge = self._aligner_edge
        clone._cell = np.array(self._cell)
        return clone

    def __str__(self):
        return (
            f"Vertex(id={self._id}, "
            f"position={self._position.tolist()}, "
            f"aligner_edge={self._aligner_edge})"
        )


class LinearVertex(_CofVertex):
    def place_building_block(self, building_block, edges):
        assert building_block.get_num_functional_groups() == 2, (
            f"{building_block} needs to have exactly 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )

        # Align the normal of the plane of best fit, defined by
        # all atoms in the building block, with the z axis.
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        normal = building_block.get_plane_normal()
        normal = get_acute_vector(
            reference=core_centroid - self._position,
            vector=normal,
        )
        building_block = building_block.with_rotation_between_vectors(
            start=normal,
            target=[0, 0, 1],
            origin=self._position,
        )

        # Rotate to place fg-fg vector along edge-edge vector.
        (fg,) = building_block.get_functional_groups(0)
        fg_centroid = building_block.get_centroid(fg.get_placer_ids())
        target = edges[0].get_position() - edges[1].get_position()
        target *= 1 if self._aligner_edge == 0 else -1

        building_block = building_block.with_rotation_between_vectors(
            start=fg_centroid - self._position,
            target=target,
            origin=self._position,
        )
        return building_block.get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        (fg,) = building_block.get_functional_groups(0)
        fg_position = building_block.get_centroid(fg.get_placer_ids())

        def fg_distance(edge):
            return euclidean(edge.get_position(), fg_position)

        edges = sorted(edges, key=fg_distance)
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}


class NonLinearVertex(_CofVertex):
    def place_building_block(self, building_block, edges):
        assert building_block.get_num_functional_groups() > 2, (
            f"{building_block} needs to have more than 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )
        # Sort to ensure that for two vertices, which are periodically
        # equivalent, "edges" has identical ordering. This means that
        # the aligner_edge is chosen consistently in both cases.
        edges = sorted(edges, key=lambda edge: edge.get_parent_id())

        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        normal = building_block.get_plane_normal(
            atom_ids=building_block.get_placer_ids(),
        )
        normal = get_acute_vector(
            reference=core_centroid - self._position,
            vector=normal,
        )
        building_block = building_block.with_rotation_between_vectors(
            start=normal,
            target=[0, 0, 1],
            origin=self._position,
        )
        (fg,) = building_block.get_functional_groups(0)
        fg_centroid = building_block.get_centroid(fg.get_placer_ids())
        edge_position = edges[self._aligner_edge].get_position()
        return building_block.with_rotation_to_minimize_angle(
            start=fg_centroid - self._position,
            target=edge_position - self._position,
            axis=np.array([0, 0, 1], dtype=np.float64),
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        # Sort to ensure that for two vertices, which are periodically
        # equivalent, "edges" has identical ordering. This means that
        # the aligner_edge is chosen consistently in both cases.
        edges = sorted(edges, key=lambda edge: edge.get_parent_id())

        fg_sorter = _FunctionalGroupSorter(building_block)
        edge_sorter = _EdgeSorter(
            edges=edges,
            aligner_edge=edges[self._aligner_edge],
            axis=fg_sorter.get_axis(),
        )
        return {
            fg_id: edge.get_id()
            for fg_id, edge in zip(
                fg_sorter.get_items(),
                edge_sorter.get_items(),
            )
        }


class UnaligningVertex(_CofVertex):
    """
    Just places a building block, does not align.

    """

    def place_building_block(self, building_block, edges):
        return building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}

    @classmethod
    def init_at_center(
        cls,
        id,
        vertices,
        aligner_edge=0,
        cell=(0, 0, 0),
    ):
        vertex = cls.__new__(cls)
        vertex._id = id
        vertex._position = sum(
            vertex.get_position() for vertex in vertices
        ) / len(vertices)
        vertex._cell = np.array(cell)
        vertex._aligner_edge = aligner_edge
        return vertex

    @classmethod
    def init_at_shifted_center(
        cls,
        id,
        vertices,
        cell_shifts,
        lattice_constants,
        aligner_edge=0,
        cell=(0, 0, 0),
    ):
        new_vertex = cls.__new__(cls)
        new_vertex._id = id

        positions = []
        for vertex, cell_shift in zip(vertices, cell_shifts):
            shift = sum(
                dim_shift * constant
                for dim_shift, constant in zip(cell_shift, lattice_constants)
            )
            positions.append(vertex.get_position() + shift)

        new_vertex._position = np.divide(
            np.sum(positions, axis=0),
            len(positions),
        )
        new_vertex._cell = np.array(cell)
        new_vertex._aligner_edge = aligner_edge
        return new_vertex



import pytest
import stk

from ....case_data import CaseData
from ...building_blocks import get_linker, get_pd_atom


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.cage.M2L4Lantern(
                    building_blocks={
                        get_pd_atom(): range(2),
                        get_linker(): range(2, 6),
                    },
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C([H])C2=C([H])C(=C1[H])C1=C([H])C([H])=N(->["
                "Pd+2]34<-N5=C([H])C([H])=C(C([H])=C5[H])C5=C([H])C("
                "[H])=C([H])C(=C5[H])C5=C([H])C([H])=N(->[Pd+2](<-N6"
                "=C([H])C([H])=C2C([H])=C6[H])(<-N2=C([H])C([H])=C(C"
                "([H])=C2[H])C2=C([H])C(=C([H])C([H])=C2[H])C2=C([H]"
                ")C([H])=N->3C([H])=C2[H])<-N2=C([H])C([H])=C(C([H])"
                "=C2[H])C2=C([H])C(=C([H])C([H])=C2[H])C2=C([H])C([H"
                "])=N->4C([H])=C2[H])C([H])=C5[H])C([H])=C1[H]"
            ),
            name=name,
        ),
    ),
)
def metal_cage_m2l4_lantern(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
Edge
====

"""

import numpy as np


class Edge:
    """
    Represents an edge in a topology graph.

    """

    def __init__(
        self,
        id,
        vertex1,
        vertex2,
        periodicity=(0, 0, 0),
        position=None,
    ):
        """
        Initialize an :class:`.Edge` instance.

        Parameters
        ----------
        id : :class:`int`
            The id of the edge.

        vertex1 : :class:`.Vertex`
            The first vertex the edge is connected to.

        vertex2 : :class:`.Vertex`
            The second vertex the edge is connected to.

        periodicity : :class:`tuple` of :class:`int`, optional
            The periodicity of the edge, when going from `vertex1` to
            `vertex2`. For example, if ``(0, 0, 0)`` the edge is not
            periodic, if ``(1, 0, -1)`` the edge is periodic going in
            the positive direction along the x axis, is not periodic
            across the y axis and is periodic in the negative direction
            along the z axis.

        position : :class:`numpy.ndarray`, optional
            The position of the edge, if ``None``, the midpoint of the
            vertices is used.

        """

        if position is None:
            self._position = (
                vertex1.get_position() + vertex2.get_position()
            ) / 2
        else:
            self._position = np.array(position, dtype=np.float64)

        self._id = id
        self._vertex1_id = vertex1.get_id()
        self._vertex2_id = vertex2.get_id()
        self._periodicity = periodicity

    def get_id(self):
        """
        Get the id of the edge.

        Returns
        -------
        :class:`int`
            The id.

        """

        return self._id

    def get_periodicity(self):
        """
        Get the periodicity of the edge.

        Returns
        -------
        :class:`tuple` of :class:`int`
            The periodicity of the edge. For example, if ``(0, 0, 0)``
            the edge is not periodic, if ``(1, 0, -1)`` the edge is
            periodic going in the positive direction along the x axis,
            is not periodic across the y axis and is periodic in the
            negative direction along the z axis.

        """

        return self._periodicity

    def is_periodic(self):
        """
        Return ``True`` if periodic.

        Returns
        -------
        :class:`bool`
            ``True`` if periodic.

        """

        return any(i != 0 for i in self._periodicity)

    def _with_scale(self, scale):
        """
        Modify the edge.

        """

        self._position *= scale
        return self

    def with_scale(self, scale):
        """
        Return a clone with a scaled position.

        Parameters
        ----------
        scale : :class:`float` or :class:`tuple` of :class:`float`
            The value by which the position of
            the :class:`.Edge` is scaled. Can be a single number if all
            axes are scaled by the same amount or a :class:`tuple` of
            three numbers if each axis is scaled by a different value.

        Returns
        -------
        :class:`.Edge`
            The clone. Has the same type as the original edge.


        """

        return self.clone()._with_scale(scale)

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.Edge`
            The clone. Has the same type as the original edge.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._id = self._id
        clone._vertex1_id = self._vertex1_id
        clone._vertex2_id = self._vertex2_id
        clone._position = np.array(self._position)
        clone._periodicity = self._periodicity
        return clone

    def get_vertex_ids(self):
        """
        Yield the ids of the vertices connected by the edge.

        The id of the first vertex is yielded first, followed by the id
        of the second vertex.

        Yields
        ------
        :class:`int`
            The id of a :class:`.Vertex`.

        """

        yield self._vertex1_id
        yield self._vertex2_id

    def get_vertex1_id(self):
        """
        Get the id of the first vertex.

        Returns
        -------
        :class:`int`
            The id of the first vertex.

        """

        return self._vertex1_id

    def get_vertex2_id(self):
        """
        Get the id of the second vertex.

        Returns
        -------
        :class:`int`
            The id of the second vertex.

        """

        return self._vertex2_id

    def get_position(self):
        """
        Get the position.

        Returns
        -------
        :class:`numpy.ndarray`
            The position of the :class:`Edge`.

        """

        return np.array(self._position)

    def _with_position(self, position):
        """
        Modify the edge.

        """

        self._position = np.array(position, dtype=np.float64)
        return self

    def with_position(self, position):
        """
        Return a clone at `position`.

        Parameters
        ----------
        position : :class:`numpy.ndarray`
            The desired position of the clone.

        Returns
        -------
        :class:`.Edge`
            The clone. Has the same type as the original edge.

        """

        return self.clone()._with_position(position)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Edge({self._id}, {self._vertex1_id}, {self._vertex2_id})"



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def single_atom(get_atom_ids):
    (a,) = get_atom_ids(1)
    return _single_atom(stk.N(a))


def _single_atom(atom):
    atoms = (atom,)
    deleters = ()
    return GenericCaseData(
        functional_group=stk.SingleAtom(atom=atom),
        atoms=atoms,
        bonders=atoms,
        deleters=deleters,
    )



import numpy as np
import pytest


@pytest.fixture(
    params=[
        -np.pi / 2,
        np.pi / 2,
    ],
)
def angle(request):
    return request.param


@pytest.fixture(
    params=[
        np.array([0, 1, 0]),
        np.array([1, 0, 0]),
        np.array([1 / np.sqrt(3), 1 / np.sqrt(3), 1 / np.sqrt(3)]),
    ],
)
def axis(request):
    return np.array(request.param)



"""
Reactions Summary
=================

"""

from typing import NamedTuple

from .atom_batch import _AtomBatch
from .bond_batch import _BondBatch


class _BondId(NamedTuple):
    """
    Identifies a bond in a molecule.

    Attributes
    ----------
    atom1_id
        The id of the first :class:`.Atom` in the bond.

    atom2_id
        The id of the sceond :class:`.Atom` in the bond.

    """

    atom1_id: int
    atom2_id: int


class _ReactionsSummary:
    """
    A summary of reaction results.

    """

    __slots__ = [
        "_num_atoms",
        "_atoms",
        "_atom_infos",
        "_positions",
        "_bonds",
        "_bond_infos",
        "_deleted_atom_ids",
        "_deleted_bond_ids",
    ]

    def __init__(self, num_atoms, reaction_results):
        """
        Initialize a :class:`.ReactionsSummary` instance.

        Parameters
        ----------
        num_atoms : :class:`int`
            The number of atoms in molecule being constructed,
            before this summary is taken into account.

        reaction_results : :class:`iterable`
            Holds the :class:`.ReactionResult` instances to be
            summarized.

        """

        # This will get updated as reaction results are added to the
        # summary.
        self._num_atoms = num_atoms
        self._atoms = []
        self._atom_infos = []
        self._positions = []
        self._bonds = []
        self._bond_infos = []
        self._deleted_atom_ids = set()
        self._deleted_bond_ids = set()

        for result in reaction_results:
            self._with_reaction_result(result)
            self._num_atoms += len(result.get_new_atoms())

    def _with_reaction_result(self, result):
        """
        Add the `result` to the summary.

        Parameters
        ----------
        result : :class:`.ReactionResult`
            The result to add to the summary.

        Returns
        -------
        None : :class:`NoneType`

        """

        atom_batch = _AtomBatch(
            atoms=result.get_new_atoms(),
            num_atoms=self._num_atoms,
        )
        self._with_atom_batch(atom_batch)

        bond_batch = _BondBatch(
            bonds=result.get_new_bonds(),
            atom_map=atom_batch.get_atom_map(),
        )
        self._with_bond_batch(bond_batch)

        self._deleted_atom_ids.update(
            atom.get_id() for atom in result.get_deleted_atoms()
        )

        self._deleted_bond_ids.update(
            _BondId(
                atom1_id=bond.get_atom1().get_id(),
                atom2_id=bond.get_atom2().get_id(),
            )
            for bond in result.get_deleted_bonds()
        )

    def _with_atom_batch(self, batch):
        """
        Add a batch of atoms to the summary.

        Parameters
        ----------
        batch : :class:`._AtomBatch`
            A batch of atoms.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._atoms.extend(batch.get_atoms())
        self._atom_infos.extend(batch.get_atom_infos())
        self._positions.extend(batch.get_positions())

    def _with_bond_batch(self, batch):
        """
        Add a batch of bonds to the summary.

        Parameters
        ----------
        batch : :class:`.BondBatch`
            A batch of bonds.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._bonds.extend(batch.get_bonds())
        self._bond_infos.extend(batch.get_bond_infos())

    def get_atoms(self):
        """
        Yield the atoms in the summary.

        Yields
        ------
        :class:`.Atom`
            An atom.

        """

        yield from self._atoms

    def get_atom_infos(self):
        """
        Yield infos about atoms in the summary.

        Yields
        ------
        :class:`.AtomInfo`
            Info about an atom.

        """

        yield from self._atom_infos

    def get_bonds(self):
        """
        Yield the bonds in the summary.

        Yields
        ------
        :class:`.Bond`
            A bond.

        """

        yield from self._bonds

    def get_bond_infos(self):
        """
        Yield infos about the bonds in the summary.

        Yields
        ------
        :class:`.BondInfo`
            Info about a bond.

        """

        yield from self._bond_infos

    def get_deleted_atom_ids(self):
        """
        Yield the ids of deletable atoms held by the summary.

        Yields
        ------
        :class:`int`
            The id of an atom which should be deleted.

        """

        yield from self._deleted_atom_ids

    def get_deleted_bond_ids(self):
        """
        Yield the atom ids of bonds to be deleted held by the summary.

        Yields
        ------
        :class:`tuple`
            A tuple of the atom ids of the bond which should be
            deleted.

        """

        yield from self._deleted_bond_ids

    def get_positions(self):
        """
        Yield the positions of atoms held by the summary.

        Yields
        ------
        :class:`numpy.ndarray`
            The position of an atom held by the summary.

        """

        yield from self._positions



import numpy as np
import stk

from tests.utilities import (
    is_equivalent_constructed_molecule,
    is_equivalent_molecule,
)


def test_from_json(case_data):
    """
    Test :meth:`.MoleculeDejsonizer.from_json`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the dejsonizer to test and the correct
        dejsonized molecule.

    Results
    -------
    None : :class:`NoneType`

    """

    _test_from_json(
        dejsonizer=case_data.dejsonizer,
        json=case_data.json,
        molecule=case_data.molecule,
    )


def _test_from_json(
    dejsonizer,
    json,
    molecule,
):
    """
    Test :meth:`.MoleculeDejsonizer.from_json`.

    Parameters
    ----------
    dejsonizer : :class:`.MoleculeDejsonizer`
        The dejsonizer to test.

    json : :class:`dict`
        The JSON to test.

    molecule : :class:`.Molecule`
        The correct dejsonized molecule.

    Returns
    -------
    None : :class:`NoneType`

    """

    result = dejsonizer.from_json(json)

    {
        stk.Molecule: is_equivalent_molecule,
        stk.BuildingBlock: is_equivalent_molecule,
        stk.ConstructedMolecule: is_equivalent_constructed_molecule,
    }[type(molecule)](result, molecule)

    assert np.all(
        np.equal(
            molecule.get_position_matrix(),
            result.get_position_matrix(),
        )
    )



import numpy as np


def check_position_matrix(old_state, new_state, placement_results):
    assert np.all(
        np.equal(
            get_expected_position_matrix(old_state, placement_results),
            new_state.get_position_matrix(),
        )
    )


def get_expected_position_matrix(old_state, placement_results):
    return np.vstack(
        [
            old_state.get_position_matrix(),
            *(result.position_matrix for result in placement_results),
        ]
    )



"""
Vertex
======

"""

from __future__ import annotations

import typing

import numpy as np

from stk._internal.building_block import BuildingBlock

from .edge import Edge

_VertexT = typing.TypeVar("_VertexT", bound="Vertex")


class Vertex:
    """
    An abstract base class for :class:`.TopologyGraph` vertices.

    Notes:

        You might notice that some of the public methods of this
        abstract base class are implemented. This is purely for
        convenience when implementing subclasses. The implemented
        public methods are simply default implementations, which can
        be safely ignored or overridden, when implementing subclasses.
        Any private methods are implementation details of these
        default implementations.

    """

    def __init__(
        self,
        id: int,
        position: typing.Union[tuple[float, float, float], np.ndarray],
    ) -> None:
        """
        Initialize a :class:`.Vertex`.

        Parameters:

            id:
                The id of the vertex.

            position:
                The position of the vertex.

        """

        self._id = id
        self._position = np.array(position, dtype=np.float64)

    def get_id(self) -> int:
        """
        Get the id.

        Returns:

            The id.

        """

        return self._id

    def _with_scale(
        self: _VertexT,
        scale: typing.Union[float, tuple[float, float, float]],
    ) -> _VertexT:
        """
        Modify the vertex.

        """

        self._position *= scale
        return self

    def with_scale(
        self,
        scale: typing.Union[float, tuple[float, float, float]],
    ) -> Vertex:
        """
        Get a clone with a scaled position.

        Parameters:

            scale:
                The value by which the position of the :class:`Vertex`
                is scaled. Can be a single number if all axes are
                scaled by the same amount or a :class:`tuple` of three
                numbers if each axis is scaled by a different value.

        Returns:

            The clone.

        """

        return self.clone()._with_scale(scale)

    def _clone(self: _VertexT) -> _VertexT:
        clone = self.__class__.__new__(self.__class__)
        Vertex.__init__(clone, self._id, self._position)
        return clone

    def clone(self) -> Vertex:
        """
        Return a clone.

        Returns:

            The clone.

        """

        return self._clone()

    def get_position(self) -> np.ndarray:
        """
        Get the position.

        Returns:

            The position of the :class:`Vertex`.

        """

        return np.array(self._position)

    def _with_position(
        self: _VertexT,
        position: np.ndarray,
    ) -> _VertexT:
        """
        Modify the vertex.

        """

        self._position = np.array(position, dtype=np.float64)
        return self

    def with_position(
        self,
        position: np.ndarray,
    ) -> Vertex:
        """
        Get a clone at a certain position.

        Parameters:

            position:
                The desired position of the clone.

        Returns:

            The clone.

        """

        return self.clone()._with_position(position)

    def get_cell(self) -> np.ndarray:
        """
        Get the cell of the lattice in which the vertex is found.

        Returns:

            The cell of the lattice in which the vertex is found.

        """

        return np.array([0, 0, 0])

    def place_building_block(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> np.ndarray:
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters:

            building_block:
                The building block molecule which is to be placed on
                the vertex.

            edges:
                The edges to which the vertex is attached.

        Returns:

            The position matrix of `building_block` after being
            placed.

        """

        raise NotImplementedError()

    def map_functional_groups_to_edges(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> dict[int, int]:
        """
        Map functional groups to edges.

        Each functional group in `building_block` needs to be assigned
        to an edge in `edges`.

        Parameters:

            building_block : :class:`.BuildingBlock`
                The building block which is needs to have functional
                groups assigned to edges.

            edges:
                The edges to which the vertex is attached.

        Returns:

            A mapping from the id of a functional group in
            `building_block` to the id of the edge in `edges` it
            is assigned to.

        """

        raise NotImplementedError()

    def __str__(self) -> str:
        position = self._position.tolist()
        return f"Vertex(id={self._id}, position={position})"

    def __repr__(self) -> str:
        return str(self)



import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures need to be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("linear"),
        lazy_fixture("cof"),
    ),
)
def case_data(request):
    return request.param


@pytest.fixture
def constructed_molecule(case_data):
    return case_data.constructed_molecule



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def alkyne(get_atom_ids):
    a, b, c, d = get_atom_ids(4)
    return _alkyne(stk.C(a), stk.C(b), stk.C(c), stk.C(d))


def _alkyne(carbon1, atom1, carbon2, atom2):
    bonders = (carbon1,)
    deleters = (carbon2, atom2)
    return GenericCaseData(
        functional_group=stk.Alkyne(
            carbon1=carbon1,
            atom1=atom1,
            carbon2=carbon2,
            atom2=atom2,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(carbon1, atom1, carbon2, atom2),
        bonders=bonders,
        deleters=deleters,
    )



import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures need to be visible for lazy_fixture() calls.
from .complex import *  # noqa


@pytest.fixture(
    params=(lazy_fixture("host_guest_complex"),),
)
def host_guest(request):
    return request.param



import numpy as np
import stk


def test_put_caching(mongo_client):
    database_name = "_test_put_caching"
    mongo_client.drop_database(database_name)

    database = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        jsonizer=stk.MoleculeJsonizer(
            key_makers=(stk.InchiKey(),),
        ),
    )
    molecule = stk.BuildingBlock("CCC")
    database.put(molecule)
    database.put(molecule)

    cache_info = database._put.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1

    database.put(
        molecule=molecule.with_position_matrix(
            position_matrix=np.zeros((molecule.get_num_atoms(), 3)),
        ),
    )
    cache_info = database._put.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 2


def test_get_caching(mongo_client):
    database_name = "_test_get_caching"
    mongo_client.drop_database(database_name)

    database = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        jsonizer=stk.MoleculeJsonizer(
            key_makers=(stk.InchiKey(),),
        ),
    )
    molecule = stk.BuildingBlock("CCC")
    database.put(molecule)
    database.get(
        {
            stk.InchiKey().get_key_name(): stk.InchiKey().get_key(molecule),
        }
    )
    database.get(
        {
            stk.InchiKey().get_key_name(): stk.InchiKey().get_key(molecule),
        }
    )

    cache_info = database._get.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1



import numpy as np


def is_clone(edge1, edge2):
    assert edge1.get_vertex1_id() == edge2.get_vertex1_id()
    assert edge1.get_vertex2_id() == edge2.get_vertex2_id()
    assert np.all(np.equal(edge1.get_position(), edge2.get_position()))
    assert edge1.get_periodicity() == edge2.get_periodicity()
    assert edge1.get_id() == edge2.get_id()



import itertools as it

from ..utilities import is_equivalent_functional_group


def test_get_functional_groups(case_data):
    """
    Test :meth:`.BuildingBlock.get_functional_groups`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the building block to test and the correct
        functional groups.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_functional_groups(
        building_block=case_data.building_block,
        functional_groups=case_data.functional_groups,
    )


def _test_get_functional_groups(building_block, functional_groups):
    """
    Test :meth:`.BuildingBlock.get_functional_groups`.

    Parameters
    ----------
    building_block : :class:`.BuildingBlock`
        The building block to test.

    functional_groups : :class:`tuple` of :class:`.FunctionalGroup`
        The correct functional groups.

    Returns
    -------
    None : :class:`NoneType`

    """

    for fg1, fg2 in it.zip_longest(
        building_block.get_functional_groups(),
        functional_groups,
    ):
        is_equivalent_functional_group(fg1, fg2)



import typing
from collections import abc
from dataclasses import dataclass

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import (
    ReactionFactory,
)
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex

from .vertices import (
    HeadVertex,
    LinearVertex,
    TailVertex,
    UnaligningVertex,
)


class Linear(TopologyGraph):
    """
    Represents a linear polymer topology graph.

    Building blocks with two functional groups are required, unless the
    building block's position is specified to only be at the capping
    positions.

    Examples:

        *Construction*

        Linear polymers require building blocks with two functional
        groups

        .. testcode:: construction

            import stk

            bb1 = stk.BuildingBlock('NCCN', stk.PrimaryAminoFactory())
            bb2 = stk.BuildingBlock('O=CCC=O', stk.AldehydeFactory())
            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2],
                    repeating_unit='AB',
                    num_repeating_units=4,
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock('NCCN', stk.PrimaryAminoFactory())
            bb2 = stk.BuildingBlock('O=CCC=O', stk.AldehydeFactory())
            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2],
                    repeating_unit='AB',
                    num_repeating_units=4,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        polymer.get_atoms(),
                        polymer.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in polymer.get_bonds()
                ),
            )


        *Suggested Optimization*

        For :class:`.Linear` topologies, it is recommend to use the
        :class:`.Collapser` optimizer.

        .. testcode:: suggested-optimization

            import stk

            bb1 = stk.BuildingBlock('NCCN', stk.PrimaryAminoFactory())
            bb2 = stk.BuildingBlock('O=CCC=O', stk.AldehydeFactory())

            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2],
                    repeating_unit='AB',
                    num_repeating_units=4,
                    # Setting scale_steps to False tends to lead to a
                    # better structure.
                    optimizer=stk.Collapser(scale_steps=False),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock('NCCN', stk.PrimaryAminoFactory())
            bb2 = stk.BuildingBlock('O=CCC=O', stk.AldehydeFactory())

            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2],
                    repeating_unit='AB',
                    num_repeating_units=4,
                    # Setting scale_steps to False tends to lead to a
                    # better structure.
                    optimizer=stk.Collapser(scale_steps=False),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        polymer.get_atoms(),
                        polymer.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in polymer.get_bonds()
                ),
            )

        *Construction with Capping Units*

        Building blocks with a single functional group can
        also be provided as capping units

        .. testcode:: construction-with-capping-units

            import stk

            bb1 = stk.BuildingBlock('NCC(F)N', stk.PrimaryAminoFactory())
            bb2 = stk.BuildingBlock('O=CCC=O', stk.AldehydeFactory())
            bb3 = stk.BuildingBlock('BrCCN', stk.PrimaryAminoFactory())

            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2, bb3],
                    repeating_unit='ABABC',
                    num_repeating_units=1,
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCC(F)N',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
            bb3 = stk.BuildingBlock(
                smiles='BrCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )

            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(bb1, bb2, bb3),
                    repeating_unit='ABABC',
                    num_repeating_units=1,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        polymer.get_atoms(),
                        polymer.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in polymer.get_bonds()
                ),
            )

        *Defining the Orientation of Each Building Block*

        The `orientations` parameter allows the direction of each
        building block along to the chain to be flipped

        .. testcode:: defining-the-orientation-of-each-building-block

            import stk

            bb1 = stk.BuildingBlock('O=CCC=O', stk.AldehydeFactory())
            bb2 = stk.BuildingBlock('NC(Br)CN', stk.PrimaryAminoFactory())

            p1 = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2],
                    repeating_unit='AB',
                    num_repeating_units=5,
                    orientations=[1, 0.5],
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock(
                smiles='O=CC(F)CC=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='NC(Br)CN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )

            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(bb1, bb2),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    orientations=(1, 0.5),
                    random_seed=1,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        polymer.get_atoms(),
                        polymer.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in polymer.get_bonds()
                ),
            )

        In the above example, ``bb1`` is guaranteed to be flipped,
        ``bb2`` has a 50% chance of being flipped, each time it is
        placed on a node.

        Note that whether a building block will be flipped or not
        is decided during the initialization of :class:`.Linear`

        .. testcode:: defining-the-orientation-of-each-building-block

            # chain will always construct the same polymer.
            chain = stk.polymer.Linear(
                building_blocks=[bb1, bb2],
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=[0.65, 0.45],
            )
            # p2 and p3 are guaranteed to be the same as they used the
            # same topology graph.
            p2 = stk.ConstructedMolecule(chain)
            p3 = stk.ConstructedMolecule(chain)

            # chain2 may lead to a different polymer than chain,
            # despite being initialized with the same parameters.
            chain2 = stk.polymer.Linear(
                building_blocks=[bb1, bb2],
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=[0.65, 0.45]
            )

            # p4 and p5 are guaranteed to be the same because they used
            # the same topology graph. However, they may be different
            # to p2 and p3.
            p4 = stk.ConstructedMolecule(chain2)
            p5 = stk.ConstructedMolecule(chain2)

        The `random_seed` parameter can be used to get reproducible
        results

        .. testcode:: defining-the-orientation-of-each-building-block

            # p6 and p7 are guaranteed to be the same, because chain3
            # and chain4 used the same random seed.

            chain3 = stk.polymer.Linear(
                building_blocks=[bb1, bb2],
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=[0.65, 0.45],
                random_seed=4,
            )
            p6 = stk.ConstructedMolecule(chain3)

            chain4 = stk.polymer.Linear(
                building_blocks=[bb1, bb2],
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=[0.65, 0.45],
                random_seed=4,
            )
            p7 = stk.ConstructedMolecule(chain4)

        *Using Numbers to Define the Repeating Unit*

        The repeating unit can also be specified through the indices of
        the building blocks

        .. testcode:: using-numbers-to-define-the-repeating-unit

            import stk

            bb1 = stk.BuildingBlock('NCCN', stk.PrimaryAminoFactory())
            bb2 = stk.BuildingBlock('O=CCC=O', stk.AldehydeFactory())
            bb3 = stk.BuildingBlock('NCCN', stk.PrimaryAminoFactory())

            # p1 and p2 are different ways to write the same thing.
            p1 = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2, bb3],
                    repeating_unit='ACB',
                    num_repeating_units=1,
                ),
            )
            p2 = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[bb1, bb2, bb3],
                    repeating_unit=[0, 2, 1],
                    num_repeating_units=1,
                ),
            )
    """

    def __init__(
        self,
        building_blocks: abc.Iterable[BuildingBlock],
        repeating_unit: str | abc.Iterable[int],
        num_repeating_units: int,
        orientations: abc.Iterable[float] | None = None,
        random_seed: int | np.random.Generator | None = None,
        reaction_factory: ReactionFactory = GenericReactionFactory(),
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
    ) -> None:
        """
        Parameters:

            building_blocks (list[BuildingBlock]):
                The building blocks of the polymer.

            repeating_unit (str | list[int]):
                A string specifying the repeating unit of the polymer.
                For example, ``'AB'`` or ``'ABB'``. The first building
                block passed to `building_blocks` is ``'A'`` and so on.

                The repeating unit can also be specified by the
                indices of `building_blocks`, for example ``'ABB'``
                can be written as ``[0, 1, 1]``.

            num_repeating_units:
                The number of repeating units which are used to make
                the polymer.

            orientations (list[float]):
                For each character in the repeating unit, a value
                between ``0`` and ``1`` (both inclusive) must be given
                in a :class:`tuple`. It indicates the probability that
                each monomer will have its orientation along the chain
                flipped. If ``0`` then the monomer is guaranteed not to
                flip. If ``1`` it is guaranteed to flip. This allows
                the user to create head-to-head or head-to-tail chains,
                as well as chain with a preference for head-to-head or
                head-to-tail if a number between ``0`` and ``1`` is
                chosen. If ``None`` then ``0`` is picked in all cases.

                It is also possible to supply an orientation for every
                vertex in the final topology graph. In this case, the
                length of `orientations` must be equal to
                ``len(repeating_unit)*num_repeating_units``.

                If there is only one building block in the constructed
                polymer i.e. the `repeating_unit` has a length of 1 and
                `num_repeating_units` is 1, the building block will not
                be re-oriented, even if you provide a value to
                `orientations`.

            random_seed:
                The random seed to use when choosing random
                orientations.

            reaction_factory:
                The factory to use for creating reactions between
                functional groups of building blocks.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

        Raises:

            :class:`ValueError`
                If the length of `orientations` is not equal in length
                to `repeating_unit` or to the total number of vertices.

        """
        self._repr = (
            f"Linear({building_blocks!r}, "
            f"{repeating_unit!r}, {num_repeating_units!r})"
        )

        if not isinstance(repeating_unit, str):
            repeating_unit = tuple(repeating_unit)

        if orientations is None:
            orientations = tuple(0.0 for _ in range(len(repeating_unit)))
        else:
            orientations = tuple(orientations)

        if len(orientations) == len(repeating_unit):
            orientations = orientations * num_repeating_units

        polymer_length = len(repeating_unit) * num_repeating_units
        if len(orientations) != polymer_length:
            raise ValueError(
                "The length of orientations must match either "
                "the length of repeating_unit or the "
                "total number of vertices."
            )

        # Keep these for __repr__.
        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = num_repeating_units

        try:
            head, *body, tail = orientations
            vertices_and_edges = self._get_vertices_and_edges(
                head_orientation=head,
                body_orientations=body,
                tail_orientation=tail,
                random_seed=random_seed,
            )
            vertices = vertices_and_edges.vertices
            edges = vertices_and_edges.edges

        except ValueError:
            vertices = (UnaligningVertex(0, (0.0, 0.0, 0.0), False),)
            edges = ()

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(int(v.get_flip()) for v in vertices)

        super().__init__(
            building_block_vertices=self._get_building_block_vertices(
                building_blocks=tuple(building_blocks),
                vertices=vertices,
            ),
            edges=edges,
            reaction_factory=reaction_factory,
            construction_stages=(),
            optimizer=optimizer,
            num_processes=num_processes,
        )

    @staticmethod
    def _get_vertices_and_edges(
        head_orientation: float,
        body_orientations: abc.Iterable[float],
        tail_orientation: float,
        random_seed: int | np.random.Generator | None,
    ) -> "_VerticesAndEdges":
        """
        Get the vertices and edges of the topology graph.

        Parameters:

            head_orientation:
                The probability that the head vertex will do flipping.

            body_orientations:
                For each body vertex, the probability that it will do
                flipping.

            tail_orientation:
                The probability that the tail vertex will do flipping.

            random_seed:
                The random seed to use.

        Returns:

            The vertices and edges of the topology graph.

        """

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        choices = [True, False]
        vertices: list[LinearVertex] = [
            HeadVertex(
                id=0,
                position=np.array([0, 0, 0]),
                flip=random_seed.choice(
                    a=choices,
                    p=[head_orientation, 1 - head_orientation],
                ),
            ),
        ]
        edges: list[Edge] = []
        for i, p in enumerate(body_orientations, 1):
            flip = random_seed.choice(choices, p=[p, 1 - p])
            vertices.append(
                LinearVertex(i, np.array([i, 0, 0]), flip),
            )
            edges.append(Edge(len(edges), vertices[i - 1], vertices[i]))

        vertices.append(
            TailVertex(
                id=len(vertices),
                position=np.array([len(vertices), 0, 0]),
                flip=random_seed.choice(
                    a=choices,
                    p=[tail_orientation, 1 - tail_orientation],
                ),
            ),
        )

        edges.append(Edge(len(edges), vertices[-2], vertices[-1]))

        return _VerticesAndEdges(
            vertices=tuple(vertices),
            edges=tuple(edges),
        )

    def clone(self) -> typing.Self:
        clone = self._clone()
        clone._repr = self._repr
        clone._repeating_unit = self._repeating_unit
        clone._num_repeating_units = self._num_repeating_units
        clone._orientations = self._orientations
        return clone

    @staticmethod
    def _normalize_repeating_unit(
        repeating_unit: typing.Union[str, tuple[int, ...]],
    ) -> tuple[int, ...]:
        if isinstance(repeating_unit, tuple):
            return repeating_unit

        base = ord("A")
        return tuple(ord(letter) - base for letter in repeating_unit)

    def _get_building_block_vertices(
        self,
        building_blocks: tuple[BuildingBlock, ...],
        vertices: tuple[LinearVertex, ...],
    ) -> dict[BuildingBlock, abc.Sequence[Vertex]]:
        polymer = self._repeating_unit * self._num_repeating_units

        building_block_vertices: dict[BuildingBlock, list[LinearVertex]] = {}

        for bb_index, vertex in zip(polymer, vertices):
            bb = building_blocks[bb_index]
            building_block_vertices[bb] = building_block_vertices.get(bb, [])
            building_block_vertices[bb].append(vertex)

        return self._with_unaligning_vertices(
            building_block_vertices=building_block_vertices,
        )

    @staticmethod
    def _with_unaligning_vertices(
        building_block_vertices: dict[BuildingBlock, list[LinearVertex]],
    ) -> dict[BuildingBlock, abc.Sequence[Vertex]]:
        clone: dict[BuildingBlock, abc.Sequence[Vertex]]
        clone = {}
        terminal_ids = {
            0,
            max(
                vertex.get_id()
                for vertex_list in building_block_vertices.values()
                for vertex in vertex_list
            ),
        }
        for (
            building_block,
            vertices,
        ) in building_block_vertices.items():
            # Building blocks with 1 placer, cannot be aligned on
            # linear vertices and must therefore use an
            # UnaligningVertex. Building blocks with 1 placer can be
            # placed on terminal vertices (HeadVertex or TailVertex).
            # This can be discerned based on the knowledge that the
            # first and last vertex are the Head and Tail,
            # respectively.
            if building_block.get_num_placers() == 1:
                clone[building_block] = tuple(
                    (
                        UnaligningVertex(
                            id=vertex.get_id(),
                            position=vertex.get_position(),
                            flip=vertex.get_flip(),
                        )
                        if vertex.get_id() not in terminal_ids
                        else vertex
                    )
                    for vertex in vertices
                )
            else:
                clone[building_block] = vertices

        return clone

    def _get_scale(
        self,
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
    ) -> float:
        return max(bb.get_maximum_diameter() for bb in building_block_vertices)

    def with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> typing.Self:
        return self.clone()._with_building_blocks(building_block_map)

    def __repr__(self) -> str:
        return self._repr


@dataclass(frozen=True)
class _VerticesAndEdges:
    vertices: tuple[LinearVertex, ...]
    edges: tuple[Edge, ...]



from ..utilities import (
    are_equivalent_functional_groups,
    has_same_structure,
    is_equivalent_building_block,
    is_equivalent_molecule,
)


def test_with_functional_groups(building_block, get_functional_groups):
    """
    Test :meth:`.BuildingBlock.with_functional_groups`.

    Parameters
    ----------
    building_block : :class:`.BuildingBlock`
        The building block to test.

    get_functional_groups : :class:`callable`
        Takes a single parameter, `building_block` and returns the
        `functional_groups` parameter to use for this test.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save clone to check immutability.
    clone = building_block.clone()
    _test_with_functional_groups(
        building_block=building_block,
        functional_groups=tuple(get_functional_groups(building_block)),
    )
    is_equivalent_building_block(building_block, clone)
    has_same_structure(building_block, clone)


def _test_with_functional_groups(building_block, functional_groups):
    """
    Test :meth:`.BuildingBlock.with_functional_groups`.

    Parameters
    ----------
    building_block : :class:`.BuildingBlock`
        The building block to test.

    functional_groups : :class:`tuple` of :class:`.FunctionalGroup`
        The functional groups the new building block should hold.

    Returns
    -------
    None : :class:`NoneType`

    """

    new = building_block.with_functional_groups(functional_groups)
    are_equivalent_functional_groups(
        new.get_functional_groups(),
        functional_groups,
    )
    is_equivalent_molecule(building_block, new)
    has_same_structure(building_block, new)



class CaseData:
    """
    A test case.

    Attributes
    ----------
    building_block : :class:`.BuildingBlock`
        The building block to test.

    functional_groups : :class:`tuple` of :class:`.FunctionalGroup`
        The functional groups :attr:`.building_block` should be
        holding.

    core_atom_ids : :class:`tuple` of :class:`int`
        The correct core atom ids for :attr:`.building_block`.

    placer_ids : :class:`tuple` of :class:`int`
        The correct *placer* ids for :attr:`.building_block`.

    """

    def __init__(
        self,
        building_block,
        functional_groups,
        core_atom_ids,
        placer_ids,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block to test.

        functional_groups : :class:`tuple` of :class:`.FunctionalGroup`
            The functional groups `building_block` should be holding.

        core_atom_ids : :class:`tuple` of :class:`int`
            The correct core atom ids for `building_block`.

        placer_ids : :class:`tuple` of :class:`int`
            The correct *placer* ids for `building_block`.

        """

        self.building_block = building_block
        self.functional_groups = functional_groups
        self.core_atom_ids = core_atom_ids
        self.placer_ids = placer_ids



import pytest
import stk

from .case_data import CaseData


@pytest.fixture
def case_data(id, vertex1, vertex2, periodicity):
    return CaseData(
        edge=stk.Edge(
            id=id,
            vertex1=vertex1,
            vertex2=vertex2,
            periodicity=periodicity,
        ),
        id=id,
        vertex1_id=vertex1.get_id(),
        vertex2_id=vertex2.get_id(),
        periodicity=periodicity,
        is_periodic=any(i != 0 for i in periodicity),
    )


@pytest.fixture
def edge(case_data):
    return case_data.edge


@pytest.fixture(
    params=(
        lambda: stk.Vertex(0, [1, 2, 3]),
        lambda: stk.Vertex(3, [4, 2, 1]),
    ),
)
def vertex(request) -> stk.Vertex:
    return request.param()


@pytest.fixture
def vertex1(vertex):
    return vertex.clone()


@pytest.fixture
def vertex2(vertex):
    return vertex.clone()


@pytest.fixture(
    params=(
        (0, 0, 0),
        (0, 1, 0),
        (-1, 1, 1),
        (20, -1, 12),
    ),
)
def periodicity(request):
    return request.param


@pytest.fixture(params=(0, 12))
def id(request):
    return request.param



def check_edges(old_state, new_state):
    assert old_state.get_num_edges() == new_state.get_num_edges()
    for edge_id in range(old_state.get_num_edges()):
        assert old_state.get_edge(edge_id) is new_state.get_edge(edge_id)



"""
M3L3 Triangle
=============

"""

import numpy as np

from stk._internal.optimizers.null import NullOptimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import AngledVertex, LinearVertex


class M3L3Triangle(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(2)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=NC=C2)C3=CC=NC=C3'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M3L3Triangle(
                corners=bb1,
                linkers=bb2,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(2)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=NC=C2)C3=CC=NC=C3'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M3L3Triangle(
                corners=bb1,
                linkers=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Both `corner` and `linker` vertices require building blocks with
    two functional groups for this topology. This class replaces the
    `building_blocks` parameter with the `corner` and `linker`
    parameters.

    See :class:`.Cage` for more details and examples.

    """

    def __init__(
        self,
        corners,
        linkers,
        vertex_alignments=None,
        reaction_factory=GenericReactionFactory(),
        num_processes=1,
        optimizer=NullOptimizer(),
    ):
        """
        Initialize a :class:`.M3L3Triangle`.

        Parameters
        ----------
        corners : :class:`dict` or :class:`.BuildingBlock`
            Can be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on.

            Can also be a :class:`.BuildingBlock` instance, which
            should be placed on all corner vertices on the topology
            graph.

        linkers : :class:`dict` or :class:`.BuildingBlock`
            Can be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on.

            Can also be a :class:`.BuildingBlock` instance, which
            should be placed on all linker vertices on the topology
            graph.

        vertex_alignments : :class:`dict`, optional
            A mapping from the id of a :class:`.Vertex`
            to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is referred to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        reaction_factory : :class:`.ReactionFactory`, optional
            The reaction factory to use for creating bonds between
            building blocks.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        optimizer : :class:`.Optimizer`, optional
            Used to optimize the structure of the constructed
            molecule.

        """

        if isinstance(corners, dict):
            building_blocks = corners
        else:
            building_blocks = {corners: (0, 1, 2)}

        if isinstance(linkers, dict):
            linkers_dict = linkers
        else:
            linkers_dict = {linkers: (3, 4, 5)}

        building_blocks.update(
            (building_block, vertices)
            for building_block, vertices in linkers_dict.items()
        )

        super().__init__(
            building_blocks,
            vertex_alignments=vertex_alignments,
            reaction_factory=reaction_factory,
            num_processes=num_processes,
            optimizer=optimizer,
        )

    _x = 2 * np.sqrt(3) / 4
    _y = 2
    _angled_vertices = (
        AngledVertex(0, [0, _x, 0]),
        AngledVertex(1, [_y / 2, -_x, 0]),
        AngledVertex(2, [-_y / 2, -_x, 0]),
    )

    _vertex_prototypes = (
        *_angled_vertices,
        LinearVertex.init_at_center(
            id=3,
            vertices=(_angled_vertices[0], _angled_vertices[1]),
        ),
        LinearVertex.init_at_center(
            id=4,
            vertices=(_angled_vertices[1], _angled_vertices[2]),
        ),
        LinearVertex.init_at_center(
            id=5,
            vertices=(_angled_vertices[2], _angled_vertices[0]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(1, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(2, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[2], _vertex_prototypes[4]),
        Edge(4, _vertex_prototypes[2], _vertex_prototypes[5]),
        Edge(5, _vertex_prototypes[0], _vertex_prototypes[5]),
    )

    _num_windows = 1
    _num_window_types = 1



class CaseData:
    """
    A test case.

    Attributes
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    id : :class:`int`
        The correct id of :attr:`.vertex`.

    position : :class:`numpy.ndarray`
        The correct position of :attr:`.vertex`.

    cell : :class:`numpy.ndarray`
        The correct position of :attr:`.vertex`.

    """

    def __init__(self, vertex, id, position, cell):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        vertex : :class:`.Vertex`
            The vertex to test.

        id : :class:`int`
            The correct id of :attr:`.vertex`.

        position : :class:`numpy.ndarray`
            The correct position of :attr:`.vertex`.

        cell : :class:`numpy.ndarray`
            The correct position of :attr:`.vertex`.

        """

        self.vertex = vertex
        self.id = id
        self.position = position
        self.cell = cell



from __future__ import annotations

from typing import Callable

import stk


def test_with_ids(
    bond: stk.Bond,
    get_id_map: Callable[[stk.Bond], dict[int, int]],
) -> None:
    """
    Test :meth:`.Bond.with_ids`.

    Parameters:

        bond : :class:`.Bond`
            The bond to test.

        get_id_map:
            Takes a single parameter, `bond`, and returns a valid
            `id_map` parameter for its :meth:`.Bond.with_ids`
            method. This allows the testing of different values of
            this parameter.

    """

    id_map = get_id_map(bond)
    clone = bond.with_ids(id_map)
    assert clone is not bond

    expected_id1 = id_map.get(
        bond.get_atom1().get_id(),
        bond.get_atom1().get_id(),
    )
    assert clone.get_atom1().get_id() == expected_id1

    expected_id2 = id_map.get(
        bond.get_atom2().get_id(),
        bond.get_atom2().get_id(),
    )
    assert clone.get_atom2().get_id() == expected_id2

    assert bond.get_periodicity() == clone.get_periodicity()
    assert bond.get_order() == clone.get_order()



import pytest
import stk

from ..case_data import CaseData


@pytest.fixture(
    params=[
        "building_block.mol",
    ],
)
def path(tmpdir, request):
    return str(tmpdir / request.param)


class InitFromFileData:
    """
    Stores data for the :meth:`.BuildingBlock.init_from_file`.

    Attributes
    ----------
    building_block : :class:`.BuildingBlock`
        The building block, which will be written to a file, so that it
        can be initialized from it.

    init_functional_groups : :class:`iterable`
        Passed to the `functional_groups` parameter of
        :meth:`.BuildingBlock.init_from_file`.

    init_placer_ids : :class:`tuple` or :class:`NoneType`
        Passed to the `placer_ids` parameter of
        :meth:`.BuildingBlock.init_from_file`.

    case_data_functional_groups : :class:`tuple`
        The functional groups the initialized building block should
        have.

    case_data_core_atom_ids : :class:`tuple` of :class:`int`
        The ids of core atoms the initialized building block should
        have.

    case_data_placer_ids : :class:`tuple` of :class:`int`
        The ids of *placer* atoms the initialized building block should
        have.

    """

    def __init__(
        self,
        building_block,
        init_functional_groups,
        init_placer_ids,
        case_data_functional_groups,
        case_data_core_atom_ids,
        case_data_placer_ids,
    ):
        """
        Initialize a :class:`.InitFromData` instance.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block, which will be written to a file, so
            that it can be initialized from it.

        init_functional_groups : :class:`iterable`
            Passed to the `functional_groups` parameter of
            :meth:`.BuildingBlock.init_from_file`.

        init_placer_ids : :class:`tuple` or :class:`NoneType`
            Passed to the `placer_ids` parameter of
            :meth:`.BuildingBlock.init_from_file`.

        case_data_functional_groups : :class:`tuple`
            The functional groups the initialized building block should
            have.

        case_data_core_atom_ids : :class:`tuple` of :class:`int`
            The ids of core atoms the initialized building block should
            have.

        case_data_placer_ids : :class:`tuple` of :class:`int`
            The ids of *placer* atoms the initialized building block
            should have.

        """

        self.building_block = building_block
        self.init_functional_groups = init_functional_groups
        self.init_placer_ids = init_placer_ids
        self.case_data_functional_groups = case_data_functional_groups
        self.case_data_core_atom_ids = case_data_core_atom_ids
        self.case_data_placer_ids = case_data_placer_ids


@pytest.fixture(
    scope="session",
    params=(
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=(),
            init_placer_ids=None,
            case_data_functional_groups=(),
            case_data_core_atom_ids=(0, 1, 2, 3),
            case_data_placer_ids=(0, 1, 2, 3),
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=[stk.BromoFactory()],
            init_placer_ids=None,
            case_data_functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            case_data_core_atom_ids=(1, 2),
            case_data_placer_ids=(1, 2),
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=(),
            init_placer_ids=(1, 2),
            case_data_functional_groups=(),
            case_data_core_atom_ids=(0, 1, 2, 3),
            case_data_placer_ids=(1, 2),
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=[stk.BromoFactory()],
            init_placer_ids=(0, 3),
            case_data_functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            case_data_core_atom_ids=(1, 2),
            case_data_placer_ids=(0, 3),
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=[stk.IodoFactory()],
            init_placer_ids=None,
            case_data_functional_groups=(),
            case_data_core_atom_ids=(0, 1, 2, 3),
            case_data_placer_ids=(0, 1, 2, 3),
        ),
    ),
)
def init_from_file_data(request) -> InitFromFileData:
    """
    A :class:`.InitFromFileData` instance.

    """

    return request.param()


@pytest.fixture
def init_from_file(path, init_from_file_data):
    """
    A :class:`.CaseData` instance.

    """

    data = init_from_file_data
    data.building_block.write(path)
    return CaseData(
        building_block=stk.BuildingBlock.init_from_file(
            path=path,
            functional_groups=data.init_functional_groups,
            placer_ids=data.init_placer_ids,
        ),
        functional_groups=data.case_data_functional_groups,
        core_atom_ids=data.case_data_core_atom_ids,
        placer_ids=data.case_data_placer_ids,
    )



def test_get_atomic_number(case_data):
    """
    Test :meth:`.Atom.get_atomic_number`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case, containing the atom to test and its correct atomic
        number.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_atomic_number(case_data.atom, case_data.atomic_number)


def _test_get_atomic_number(atom, atomic_number):
    """
    Test :meth:`.Atom.get_atomic_number`

    Parameters
    ----------
    atom : :class:`.Atom`
        The atom to test.

    atomic_number : :class:`int`
        The correct atomic number of `atom`.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert atom.get_atomic_number() == atomic_number



import numpy as np


def test_get_position(case_data):
    """
    Test :meth:`.Vertex.get_position`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the vertex to test and the correct
        position.

    Returns
    -------
    None : :class:`NoneType`


    """

    _test_get_position(case_data.vertex, case_data.position)


def _test_get_position(vertex, position):
    """
    Test :meth:`.Vertex.get_position`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    position : :class:`numpy.ndarray`
        The correct position.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert np.allclose(vertex.get_position(), position, atol=1e-15)



import stk


class CaseData:
    """
    A test case.

    Attributes
    ----------
    atom_info : :class:`.AtomInfo`
        The atom info to test.

    atom : :class:`.Atom`
        The atom, which should be held by :attr:`.atom_info`.

    building_block_atom : :class:`.Atom`
        The building block atom which should be held by
        :attr:`.atom_info`.

    building_block : :class:`.BuildingBlock`
        The building block, which should be held by :attr:`.atom_info`.

    building_block_id : :class:`int`
        The building block id, which should be held by
        :attr:`.atom_info`.

    """

    def __init__(
        self,
        atom,
        building_block_atom,
        building_block,
        building_block_id,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        atom : :class:`.Atom`
            The atom, which should be held by an :class:`.AtomInfo`.

        building_block_atom : :class:`.Atom`
            The building block atom which should be held by an
            :class:`.AtomInfo`.

        building_block : :class:`.BuildingBlock`
            The building block, which should be held by an
            :class:`.AtomInfo`.

        building_block_id : :class:`int`
            The building block id, which should be held by an
            :class:`.AtomInfo`.

        """

        self.atom_info = stk.AtomInfo(
            atom=atom,
            building_block_atom=building_block_atom,
            building_block=building_block,
            building_block_id=building_block_id,
        )
        self.atom = atom
        self.building_block_atom = building_block_atom
        self.building_block = building_block
        self.building_block_id = building_block_id



"""
Functional Group Factory Utilities
==================================

"""

import rdkit.Chem.AllChem as rdkit


def get_atom_ids(query, molecule):
    """
    Yield the ids of atoms in `molecule` which match `query`.

    Multiple substructures in `molecule` can match `query` and
    therefore each set is yielded as a group.

    Parameters
    ----------
    query : :class:`str`
        A SMARTS string used to query atoms.

    molecule : :class:`.Molecule`
        A molecule whose atoms should be queried.

    Yields
    ------
    :class:`tuple` of :class:`int`
        The ids of atoms in `molecule` which match `query`.

    """

    rdkit_mol = molecule.to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol)
    yield from rdkit_mol.GetSubstructMatches(
        query=rdkit.MolFromSmarts(query),
    )



import typing

import stk

from .case_data import CaseData


def test_caching(case_data: CaseData) -> None:
    """
    Test that a :class:`.FitnessCalculator` returns cached values.
    """

    _test_caching(
        fitness_calculator=case_data.fitness_calculator,
        record=case_data.record,
        fitness_value=case_data.fitness_value,
    )


def _test_caching(
    fitness_calculator: stk.FitnessCalculator,
    record: stk.MoleculeRecord,
    fitness_value: typing.Any,
):
    assert fitness_calculator.get_fitness_value(record) is fitness_value



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(stk.Bond(stk.C(0), stk.N(1), 1),),
)
def bond(request):
    """
    A :class:`.Bond` instance.

    """

    return request.param


@pytest.fixture(
    params=(
        stk.BuildingBlock("NCCN"),
        None,
    ),
)
def building_block(request):
    """
    A valid building block for a :class:`.BondInfo`.

    """

    return request.param


@pytest.fixture(
    params=(
        1,
        None,
    ),
)
def building_block_id(request):
    """
    A valid building block id.

    """

    return request.param


@pytest.fixture
def case_data(bond, building_block, building_block_id):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(bond, building_block, building_block_id)



"""
Generic Functional Group
========================

"""


from __future__ import annotations

import typing

from stk._internal.atom import Atom
from stk._internal.functional_groups.functional_group import FunctionalGroup

_T = typing.TypeVar("_T", bound="GenericFunctionalGroup")


class GenericFunctionalGroup(FunctionalGroup):
    """
    A functional group which defines general atomic classes.

    *Bonders* are atoms which should have bonds added by a
    :class:`.Reaction`. *Deleters* are atoms which should be removed
    by a :class:`.Reaction`.

    This interface allows the same reactions to be carried out across
    different functional groups, without knowing which specific
    elements it holds.

    """

    def __init__(
        self,
        atoms: tuple[Atom, ...],
        bonders: tuple[Atom, ...],
        deleters: tuple[Atom, ...],
        placers: typing.Optional[tuple[Atom, ...]] = None,
    ) -> None:
        """
        Initialize a :class:`.GenericFunctionalGroup`.

        Parameters:

            atoms:
                The atoms in the functional group.

            bonders:
                The bonder atoms in the functional group.

            deleters:
                The deleter atoms in the functional group.

            placers:
                The placer atoms of the functional group. If ``None``,
                the `bonders` will be used.

        """

        deleter_set = set(atom.get_id() for atom in deleters)
        super().__init__(
            atoms=atoms,
            placers=bonders if placers is None else placers,
            core_atoms=tuple(
                atom for atom in atoms if atom.get_id() not in deleter_set
            ),
        )
        self._bonders = bonders
        self._deleters = deleters

    def _clone(self: _T) -> _T:
        clone = super()._clone()
        clone._bonders = self._bonders
        clone._deleters = self._deleters
        return clone

    def clone(self) -> GenericFunctionalGroup:
        return self._clone()

    def with_atoms(
        self,
        atom_map: dict[int, Atom],
    ) -> GenericFunctionalGroup:
        return GenericFunctionalGroup(
            atoms=tuple(atom_map.get(a.get_id(), a) for a in self._atoms),
            bonders=tuple(atom_map.get(a.get_id(), a) for a in self._bonders),
            deleters=tuple(
                atom_map.get(a.get_id(), a) for a in self._deleters
            ),
            placers=tuple(atom_map.get(a.get_id(), a) for a in self._placers),
        )

    def _with_ids(self: _T, id_map: dict[int, int]) -> _T:
        super()._with_ids(id_map)
        self._bonders = tuple(
            bonder.with_id(
                id=id_map.get(
                    bonder.get_id(),
                    bonder.get_id(),
                )
            )
            for bonder in self._bonders
        )
        self._deleters = tuple(
            deleter.with_id(
                id=id_map.get(
                    deleter.get_id(),
                    deleter.get_id(),
                ),
            )
            for deleter in self._deleters
        )
        return self

    def with_ids(
        self,
        id_map: dict[int, int],
    ) -> GenericFunctionalGroup:
        return self.clone()._with_ids(id_map)

    def get_bonders(self) -> typing.Iterator[Atom]:
        """
        Yield bonder atoms in the functional group.

        These are atoms which have bonds added during
        :class:`.ConstructedMolecule` construction.

        Yields:
            A bonder atom.

        """

        yield from self._bonders

    def get_num_bonders(self) -> int:
        """
        Get the number of bonder atoms.

        Returns:
            The number of bonder atoms.

        """

        return len(self._bonders)

    def get_bonder_ids(self) -> typing.Iterator[int]:
        """
        Yield the ids of bonder atoms.

        Yields:
            The id of a bonder :class:`.Atom`.

        """

        yield from (a.get_id() for a in self._bonders)

    def get_deleters(self) -> typing.Iterator[Atom]:
        """
        Yield the deleter atoms in the functional group.

        These are atoms which are removed during
        :class:`.ConstructedMolecule` construction.

        Yields:
            A deleter atom.

        """

        yield from self._deleters

    def get_deleter_ids(self) -> typing.Iterator[int]:
        """
        Yield the ids of deleter atoms.

        Yields:
            The id of a deleter :class:`.Atom`.

        """

        yield from (a.get_id() for a in self._deleters)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"atoms={self._atoms}, "
            f"bonders={self._bonders}, "
            f"deleters={self._deleters}"
            ")"
        )



import pytest
import stk

from ..case_data import CaseData


def _get_case_data_1() -> CaseData:
    topology_graph = stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=2,
    )
    return CaseData.new(
        fitness_normalizer=stk.ReplaceFitness(
            get_replacement=lambda fitness_values: min(
                fitness_value
                for fitness_value in fitness_values.values()
                if fitness_value is not None
            )
            / 2,
            filter=lambda fitness_values, record: fitness_values[record]
            is None,
        ),
        fitness_values={
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): (1, 1),
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): (2, 2),
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): (3, 3),
            stk.MoleculeRecord(topology_graph): (None, 0.5),
        },
    )


@pytest.fixture(
    scope="session",
    params=(_get_case_data_1,),
)
def replace_fitness(request: pytest.FixtureRequest) -> CaseData:
    return request.param()



import numpy as np
import pytest
import rdkit.Chem.AllChem as rdkit
import stk

from ....case_data import CaseData


def _get_macrocycle() -> stk.BuildingBlock:
    macrocycle = stk.ConstructedMolecule(
        topology_graph=stk.macrocycle.Macrocycle(
            building_blocks=(
                stk.BuildingBlock("BrCCCBr", [stk.BromoFactory()]),
            ),
            repeating_unit="A",
            num_repeating_units=5,
        ),
    )
    return stk.BuildingBlock.init_from_molecule(macrocycle)


def _get_case_data_1() -> CaseData:
    macrocycle = _get_macrocycle()
    macrocycle_ids = tuple(
        max(
            rdkit.GetSymmSSSR(macrocycle.to_rdkit_mol()),
            # mypy throws an incorrect type error here, I believe it has
            # issues with generic functions sometimes.
            key=len,  # type: ignore
        )
    )
    return CaseData(
        vertex=stk.rotaxane.CycleVertex(0, (1, 2, 3), False),
        edges=(),
        building_block=_get_macrocycle(),
        position=np.array([1, 2, 3], dtype=np.float64),
        alignment_tests={
            get_plane_normal: np.array([1, 0, 0], dtype=np.float64),
        },
        functional_group_edges={},
        position_ids=macrocycle_ids,
    )


def _get_case_data_2() -> CaseData:
    macrocycle = _get_macrocycle()
    macrocycle_ids = tuple(
        max(
            rdkit.GetSymmSSSR(macrocycle.to_rdkit_mol()),
            # mypy throws an incorrect type error here, I believe it has
            # issues with generic functions sometimes.
            key=len,  # type: ignore
        )
    )
    return CaseData(
        vertex=stk.rotaxane.CycleVertex(0, (1, 2, 3), True),
        edges=(),
        building_block=_get_macrocycle(),
        position=np.array([1, 2, 3], dtype=np.float64),
        alignment_tests={
            get_plane_normal: np.array([-1, 0, 0], dtype=np.float64),
        },
        functional_group_edges={},
        position_ids=macrocycle_ids,
    )


def get_plane_normal(building_block):
    macrocycle = max(
        rdkit.GetSymmSSSR(building_block.to_rdkit_mol()),
        key=len,
    )
    centroid = building_block.get_centroid()
    (atom1_position,) = building_block.get_atomic_positions(0)
    normal = -stk.get_acute_vector(
        reference=atom1_position - centroid,
        vector=building_block.get_plane_normal(macrocycle),
    )
    if np.allclose(normal, [1, 0, 0], atol=1e-13):
        return np.array([1, 0, 0], dtype=np.float64)
    elif np.allclose(normal, [-1, 0, 0], atol=1e-13):
        return np.array([-1, 0, 0], dtype=np.float64)

    return normal


@pytest.fixture(
    scope="session",
    params=(
        _get_case_data_1,
        _get_case_data_2,
    ),
)
def cycle(request) -> CaseData:
    return request.param()



from typing import Any

import numpy as np
import stk

from .case_data import CaseData


def test_normalize(case_data: CaseData) -> None:
    _test_normalize(
        fitness_normalizer=case_data.fitness_normalizer,
        population=case_data.fitness_values,
        normalized=case_data.normalized,
    )


def _test_normalize(
    fitness_normalizer: stk.FitnessNormalizer[stk.MoleculeRecord[Any]],
    population: dict[stk.MoleculeRecord[Any], Any],
    normalized: dict[stk.MoleculeRecord[Any], Any],
) -> None:
    result = fitness_normalizer.normalize(population)
    assert len(result) == len(population)
    for record, fitness_value in result.items():
        if isinstance(fitness_value, np.ndarray):
            fitness_value = tuple(fitness_value)
        assert fitness_value == normalized[record]



from .above_average import *  # noqa
from .best import *  # noqa
from .filter_batches import *  # noqa
from .filter_molecules import *  # noqa
from .remove_batches import *  # noqa
from .remove_molecules import *  # noqa
from .worst import *  # noqa



import itertools as it


def test_get_building_blocks(case_data):
    """
    Test :meth:`.ConstructedMolecule.get_building_blocks`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the constructed molecule to test and the
        correct building blocks.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_building_blocks(
        constructed_molecule=case_data.constructed_molecule,
        building_blocks=case_data.building_blocks,
    )


def _test_get_building_blocks(constructed_molecule, building_blocks):
    """
    Test :meth:`.ConstructedMolecule.get_building_blocks`.

    Parameters
    ----------
    constructed_molecule : :class:`.ConstrcutedMolecule`
        The constructed molecule to test.

    building_block : :class:`tuple` of :class:`.BuildingBlock`
        The correct building blocks.

    Returns
    -------
    None : :class:`NoneType`

    """

    for building_block1, building_block2 in it.zip_longest(
        constructed_molecule.get_building_blocks(),
        building_blocks,
    ):
        assert building_block1 is building_block2



def test_get_num_atoms(case_data):
    """
    Test :meth:`.Molecule.get_num_atoms`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the number of atoms
        it should have.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.molecule.get_num_atoms() == case_data.num_atoms



def test_dump_molecular_structures(
    request,
    tmp_path,
    molecule_db,
    name_db,
    case_data,
):
    """
    Dump molecular structures.

    This test dumps molecules to files and to a MongoDB database
    so that they can be visually inspected.

    Parameters
    ----------
    request : :class:`pytest.FixtureRequest`
        Holds information about the requesting test.

    tmp_path : :class:`pathlib2.Path`
        A path into which the structure of the molecule is saved.

    molecule_db : :class:`.MoleculeDatabase`
        A database into which the structure of the molecule  is saved.

    name_db : :class:`.ValueDatabase`
        A database into which name the name of the molecule is saved.

    Returns
    -------
    None : :class:`NoneType`

    """

    case_data.molecule.write(tmp_path / f"{request.node.name}.mol")
    molecule_db.put(case_data.molecule)
    name_db.put(case_data.molecule, request.node.name)



class CaseData:
    """
    A :class:`.Reaction` test case.

    Attributes
    ----------
    reaction : :class:`.Reaction`
        The reaction to test.

    new_atoms : :class:`tuple` of :class:`.NewAtom`
        The atoms, which should be added by :attr:`reaction`.

    new_bonds : :class:`tuple` of :class:`.Bond`
        The bonds, which should be added by :attr:`reaction`.

    deleted_atoms : :class:`tuple` of :class:`.Atom`
        The atoms, which should be deleted by :attr:`reaction`.

    deleted_bonds : :class:`tuple` of :class:`.Bond`
        The bonds, which should be deleted by :attr:`reaction`.

    """

    def __init__(
        self,
        reaction,
        new_atoms,
        new_bonds,
        deleted_atoms,
        deleted_bonds,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        reaction : :class:`.Reaction`
            The reaction to test.

        new_atoms : :class:`tuple` of :class:`.NewAtom`
            The atoms, which should be added by `reaction`.

        new_bonds : :class:`tuple` of :class:`.Bond`
            The bonds, which should be added by `reaction`.

        deleted_atoms : :class:`tuple` of :class:`.Atom`
            The atoms, which should be deleted by `reaction`.

        deleted_bonds : :class:`tuple` of :class:`.Bond`
            The bonds, which should be deleted by :attr:`reaction`.

        """

        self.reaction = reaction
        self.new_atoms = new_atoms
        self.new_bonds = new_bonds
        self.deleted_atoms = deleted_atoms
        self.deleted_bonds = deleted_bonds



"""
Placements Summary
==================

"""

from collections import defaultdict
from typing import TYPE_CHECKING, Iterable

import numpy as np

from stk._internal.atom import Atom
from stk._internal.atom_info import AtomInfo
from stk._internal.bond import Bond
from stk._internal.bond_info import BondInfo
from stk._internal.functional_groups.functional_group import FunctionalGroup
from stk._internal.topology_graphs.topology_graph.utilities import (
    _PlacementResult,
)

from .atom_batch import _AtomBatch
from .bond_batch import _BondBatch

if TYPE_CHECKING:
    from stk._internal.building_block import BuildingBlock


class _PlacementsSummary:
    """
    A summary of placement results.

    """

    __slots__ = [
        "_atoms",
        "_atom_infos",
        "_bonds",
        "_bond_infos",
        "_edge_functional_groups",
        "_position_matrices",
        "_num_atoms",
    ]

    _atoms: list[Atom]
    _atom_infos: list[AtomInfo]
    _bonds: list[Bond]
    _bond_infos: list[BondInfo]
    _edge_functional_groups: defaultdict[int, list[FunctionalGroup]]
    _position_matrices: list[np.ndarray]

    def __init__(
        self,
        building_blocks: Iterable["BuildingBlock"],
        placement_results: Iterable[_PlacementResult],
        num_atoms: int,
        num_previous_placements: int,
    ) -> None:
        """
        Initialize a :class:`._PlacementsSummary` instance.

        Parameters:

            building_blocks:
                The building blocks which were placed.

            placement_results:
                Holds a :class:`_PlacementResults` instance for each
                building block in `building_blocks`.

            num_atoms:
                The number of atoms in molecule being constructed,
                before this summary is taken into account.

            num_previous_placements:
                The total number of building block placements done
                previously.

        """
        self._atoms = []
        self._atom_infos = []
        self._bonds = []
        self._bond_infos = []
        self._edge_functional_groups = defaultdict(list)
        self._position_matrices = []
        # This will get updated as placement results are added to the
        # summary.
        self._num_atoms = num_atoms

        for id_, (building_block, result) in enumerate(
            zip(building_blocks, placement_results),
            num_previous_placements,
        ):
            self._with_placement_result(building_block, id_, result)
            self._num_atoms += building_block.get_num_atoms()

    def _with_placement_result(
        self,
        building_block: "BuildingBlock",
        building_block_id: int,
        result: _PlacementResult,
    ) -> None:
        """
        Add the placement result to the summary.

        Parameters:

            building_block:
                The building block which was placed.

            building_block_id:
                A unique for the building block placement.

            result:
                The result of the placement.

        """

        self._position_matrices.append(result.position_matrix)

        atom_batch = _AtomBatch(
            atoms=building_block.get_atoms(),
            num_atoms=self._num_atoms,
            building_block=building_block,
            building_block_id=building_block_id,
        )
        self._with_atom_batch(atom_batch)
        id_map = atom_batch.get_id_map()

        bond_batch = _BondBatch(
            bonds=building_block.get_bonds(),
            id_map=id_map,
            building_block=building_block,
            building_block_id=building_block_id,
        )
        self._with_bond_batch(bond_batch)

        self._with_functional_group_edges(
            building_block=building_block,
            functional_group_edges=result.functional_group_edges,
            id_map=id_map,
        )

    def _with_atom_batch(self, batch):
        """
        Add a batch of atoms to the summary.

        Parameters
        ----------
        batch : :class:`._AtomBatch`
            A batch of atoms.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._atoms.extend(batch.get_atoms())
        self._atom_infos.extend(batch.get_atom_infos())

    def _with_bond_batch(self, batch):
        """
        Add a batch of bonds to the summary.

        Parameters
        ----------
        batch : :class:`.BondBatch`
            A batch of bonds.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._bonds.extend(batch.get_bonds())
        self._bond_infos.extend(batch.get_bond_infos())

    def _with_functional_group_edges(
        self,
        building_block: "BuildingBlock",
        functional_group_edges: dict[int, int],
        id_map: dict[int, int],
    ):
        """
        Add the mapping from functional groups to edges.

        Parameters:

            building_block:
                The building block which owns the functional groups.

            functional_group_edges:
                Maps the id of each functional group of
                `building_block` the the id of an edge.

            id_map:
                Maps the ids of atoms in `building_block` to the new
                atoms of the molecule being constructed.

        """

        functional_groups = building_block.get_functional_groups(
            fg_ids=functional_group_edges,
        )
        edge_ids = functional_group_edges.values()
        functional_group_edges_ = zip(functional_groups, edge_ids)
        for functional_group, edge_id in functional_group_edges_:
            self._edge_functional_groups[edge_id].append(
                functional_group.with_ids(id_map)
            )

    def get_atoms(self):
        """
        Yield the atoms in the summary.

        Yields
        ------
        :class:`.Atom`
            An atom.

        """

        yield from self._atoms

    def get_atom_infos(self):
        """
        Yield infos about atoms in the summary.

        Yields
        ------
        :class:`.AtomInfo`
            Info about an atom.

        """

        yield from self._atom_infos

    def get_bonds(self):
        """
        Yield the bonds in the summary.

        Yields
        ------
        :class:`.Bond`
            A bond.

        """

        yield from self._bonds

    def get_bond_infos(self):
        """
        Yield infos about the bonds in the summary.

        Yields
        ------
        :class:`.BondInfo`
            Info about a bond.

        """

        yield from self._bond_infos

    def get_position_matrix(self):
        """
        Get a position matrix for the atoms in the summary.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix.

        """

        return np.vstack(self._position_matrices)

    def get_edge_functional_groups(self):
        """
        Yield the edge ids and functional groups associated with them.

        Yields
        ------
        :class:`tuple`
            Holds the edge id as the first element, and a
            :class:`tuple` of :class:`.FunctionalGroup` instances
            as the second element.

        """

        yield from self._edge_functional_groups.items()



import numpy as np
import pytest
import stk

from ....case_data import CaseData


def get_aligned_building_block(building_block, target):
    return building_block.with_rotation_between_vectors(
        start=building_block.get_direction(),
        target=target,
        origin=building_block.get_centroid(
            atom_ids=building_block.get_placer_ids(),
        ),
    )


def get_direction(building_block):
    direction = building_block.get_direction()
    if np.allclose(direction, [1, 0, 0], atol=1e-13):
        return np.array([1, 0, 0], dtype=np.float64)
    return direction


@pytest.fixture(
    params=(
        lambda: CaseData(
            vertex=stk.host_guest.GuestVertex(
                id=0,
                position=(1, 2, 3),
                start=(1, 2, 3),
                target=(1, 0, 0),
            ),
            edges=(),
            building_block=get_aligned_building_block(
                building_block=stk.BuildingBlock("BrCCBr"),
                target=(1, 2, 3),
            ),
            position=np.array([1, 2, 3], dtype=np.float64),
            alignment_tests={
                get_direction: np.array([1, 0, 0], dtype=np.float64),
            },
            functional_group_edges={},
        ),
    ),
)
def guest(request) -> CaseData:
    return request.param()



import numpy as np


def test_write(molecule, get_position_matrix, path):
    """
    Test :meth:`.Molecule.write`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to write.

    get_position_matrix : :class:`callable
        Takes a single parameter, `molecule`, and returns a valid
        position matrix to use for this test.

    path : :class:`str`
        The path to which the test writes.

    Returns
    -------
    None : :class:`NoneType`

    """

    position_matrix = get_position_matrix(molecule)
    molecule.with_position_matrix(position_matrix).write(path)
    loaded = molecule.with_structure_from_file(path)
    assert np.allclose(
        a=position_matrix,
        b=loaded.get_position_matrix(),
        atol=1e-4,
    )



import itertools as it

import pytest
import stk

from ..case_data import CaseData
from .utilities import MockConstructionState, MockEdge


@pytest.fixture
def one_one_reaction(
    periodicity,
    functional_group1,
    functional_group1_2,
    bond_order,
):
    bond_order_key = frozenset(
        {
            type(functional_group1),
            type(functional_group1_2),
        }
    )
    edge = MockEdge(0, periodicity)
    return CaseData(
        factory=stk.GenericReactionFactory(
            bond_orders={
                bond_order_key: bond_order,
            },
        ),
        construction_state=MockConstructionState(
            edges=(edge,),
            edge_functional_groups={
                0: (
                    functional_group1,
                    functional_group1_2,
                )
            },
        ),
        edge_group=stk.EdgeGroup(edges=(edge,)),
        reaction_result=stk.ReactionResult(
            new_atoms=(),
            new_bonds=get_new_bonds(
                functional_group1=functional_group1,
                functional_group2=functional_group1_2,
                order=bond_order,
                periodicity=periodicity,
            ),
            deleted_atoms=it.chain(
                functional_group1.get_deleters(),
                functional_group1_2.get_deleters(),
            ),
            deleted_bonds=(),
        ),
    )


def get_new_bonds(
    functional_group1,
    functional_group2,
    order,
    periodicity,
):
    yield stk.Bond(
        atom1=next(functional_group1.get_bonders()),
        atom2=next(functional_group2.get_bonders()),
        order=order,
        periodicity=periodicity,
    )



import pytest
import stk

from ..case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            fitness_calculator=stk.FitnessFunction(
                fitness_function=stk.Molecule.get_num_atoms,
            ),
            molecule=stk.BuildingBlock("BrCCBr"),
            fitness_value=8,
        ),
    ),
)
def fitness_function(request) -> CaseData:
    return request.param()



import numpy as np
import pytest

from .utilities import get_maximum_diameter


@pytest.mark.parametrize(
    argnames=("position_matrix", "atom_ids", "maximum_diameter"),
    argvalues=(
        (
            np.array(
                [
                    [1.0, 2.0, 3.0],
                    [1.0, 2.0, 4.0],
                ]
            ),
            (0,),
            0,
        ),
        (
            np.array(
                [
                    [1.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0],
                    [9.0, 0.0, 0.0],
                ]
            ),
            (0, 1, 2),
            10,
        ),
        (
            np.array(
                [
                    [1.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0],
                    [9.0, 0.0, 0.0],
                ]
            ),
            (0, 2),
            8,
        ),
        (
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [3.0, 4.0, 0.0],
                ]
            ),
            (0, 1),
            5,
        ),
    ),
)
def test_get_direction_helper(
    position_matrix,
    atom_ids,
    maximum_diameter,
):
    """
    Test :func:`.get_maximum_diameter`.

    Parameters
    ----------
    position_matrix : :class:`numpy.ndarray`
        The position matrix to test.

    atom_ids : :class:`tuple` of :class:`int`
        The ids of atoms included in the test.

    maximum_diameter : :class:`float`
        The correct maximum diameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert np.allclose(
        a=get_maximum_diameter(position_matrix, atom_ids),
        b=maximum_diameter,
        atol=1e-32,
    )



"""
Linkerless Honeycomb
====================

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cof import Cof
from .vertices import NonLinearVertex


class LinkerlessHoneycomb(Cof):
    """
    Represents a honeycomb COF topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.LinkerlessHoneycomb(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='BrCC(CBr)CBr',
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                lattice_size=(3, 3, 1),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
            ),
        )

    ``Collapser(scale_steps=False)`` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.LinkerlessHoneycomb(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='BrCC(CBr)CBr',
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                lattice_size=(3, 3, 1),
                optimizer=stk.Collapser(scale_steps=False),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
            ),
        )


    Building blocks with three functional groups are required
    for this topology graph.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cof-topology-graph-examples`:
    *Multi-Building Block COF Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 1

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1.0, 0.0, 0.0]),
        np.array([0.5, 0.866, 0.0]),
        np.array([0.0, 0.0, 5 / 1.7321]),
    )

    _vertex_prototypes = (
        NonLinearVertex(0, (1 / 3) * _a + (1 / 3) * _b + (1 / 2) * _c),
        NonLinearVertex(1, (2 / 3) * _a + (2 / 3) * _b + (1 / 2) * _c),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(
            id=1,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            periodicity=(-1, 0, 0),
        ),
        Edge(
            id=2,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            periodicity=(0, -1, 0),
        ),
    )



"""
Ring Amine
==========

"""

from .functional_group import FunctionalGroup


class RingAmine(FunctionalGroup):
    """
    Represents an amine bonded to a ring.

    The structure of the functional group is given by the pseudo-SMILES
    ``[hydrogen1][nitrogen]([hydrogen2])[carbon1][carbon2]
    ([hydrogen3])[carbon3]``.

    """

    def __init__(
        self,
        nitrogen,
        hydrogen1,
        hydrogen2,
        carbon1,
        carbon2,
        hydrogen3,
        carbon3,
    ):
        """
        Initializes a :class:`.RingAmine` instance.

        Parameters
        ----------
        nitrogen : :class:`.N`
            The ``[nitrogen]`` atom.

        hydrogen1 : :class:`.H`
            The ``[hydrogen1]`` atom.

        hydrogen2 : :class:`.H`
            The ``[hydrogen2]`` atom.

        carbon1 : :class:`.C`
            The ``[carbon1]`` atom.

        carbon2 : :class:`.C`
            The ``[carbon2]`` atom.

        hydrogen3 : :class:`.H`
            The ``[hydrogen3]`` atom.

        carbon3 : :class:`.C`
            The ``[carbon3]`` atom.

        """

        self._nitrogen = nitrogen
        self._hydrogen1 = hydrogen1
        self._hydrogen2 = hydrogen2
        self._hydrogen3 = hydrogen3
        self._carbon1 = carbon1
        self._carbon2 = carbon2
        self._carbon3 = carbon3
        atoms = (
            nitrogen,
            hydrogen1,
            hydrogen2,
            carbon1,
            carbon2,
            hydrogen3,
            carbon3,
        )
        super().__init__(
            atoms=atoms,
            placers=(nitrogen,),
            core_atoms=(nitrogen,),
        )

    def get_nitrogen(self):
        """
        Get the ``[nitrogen]`` atom.

        Returns
        -------
        :class:`.N`
            The ``[nitrogen]`` atom.

        """

        return self._nitrogen

    def get_hydrogen1(self):
        """
        Get the ``[hydrogen1]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen1]`` atom.

        """

        return self._hydrogen1

    def get_hydrogen2(self):
        """
        Get the ``[hydrogen2]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen2]`` atom.

        """

        return self._hydrogen2

    def get_carbon1(self):
        """
        Get the ``[carbon1]`` atom.

        Returns
        -------
        :class:`.C`
            The ``[carbon1]`` atom.

        """

        return self._carbon1

    def get_carbon2(self):
        """
        Get the ``[carbon2]`` atom.

        Returns
        -------
        :class:`.C`
            The ``[carbon2]`` atom.

        """

        return self._carbon2

    def get_hydrogen3(self):
        """
        Get the ``[hydrogen3]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen3]`` atom.

        """

        return self._hydrogen3

    def get_carbon3(self):
        """
        Get the ``[carbon3]`` atom.

        Returns
        -------
        :class:`.C`
            The ``[carbon3]`` atom.

        """

        return self._carbon3

    def clone(self):
        clone = super().clone()
        clone._nitrogen = self._nitrogen
        clone._hydrogen1 = self._hydrogen1
        clone._hydrogen2 = self._hydrogen2
        clone._hydrogen3 = self._hydrogen3
        clone._carbon1 = self._carbon1
        clone._carbon2 = self._carbon2
        clone._carbon3 = self._carbon3
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._nitrogen = atom_map.get(
            self._nitrogen.get_id(),
            self._nitrogen,
        )
        clone._hydrogen1 = atom_map.get(
            self._hydrogen1.get_id(),
            self._hydrogen1,
        )
        clone._hydrogen2 = atom_map.get(
            self._hydrogen2.get_id(),
            self._hydrogen2,
        )
        clone._hydrogen3 = atom_map.get(
            self._hydrogen3.get_id(),
            self._hydrogen3,
        )
        clone._carbon1 = atom_map.get(
            self._carbon1.get_id(),
            self._carbon1,
        )
        clone._carbon2 = atom_map.get(
            self._carbon2.get_id(),
            self._carbon2,
        )
        clone._carbon3 = atom_map.get(
            self._carbon3.get_id(),
            self._carbon3,
        )
        return clone

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self._nitrogen}, {self._hydrogen1}, {self._hydrogen2}, "
            f"{self._carbon1}, {self._carbon2}, {self._hydrogen3}, "
            f"{self._carbon3})"
        )



import typing
from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from stk._internal.ea.crossover.record import CrossoverRecord
from stk._internal.ea.mutation.record import MutationRecord


@dataclass(frozen=True, slots=True)
class FitnessValues:
    """
    Fitness values of a molecule.

    Parameters:
        raw:
            Fitness value before normalization.
        normalized:
            Fitness value after normalization.
    """

    raw: typing.Any
    """Fitness value before normalization."""
    normalized: float
    """Fitness value after normalization."""


T = typing.TypeVar("T")


class Generation(typing.Generic[T]):
    """
    An abstract base class for EA generations.

    Notes:
        You might notice that the public methods of this abstract base
        class are implemented. This is just a default implementation, which
        you can ignore or override when implementing subclasses.
    """

    def __init__(
        self,
        fitness_values: "dict[T, FitnessValues]",
        mutation_records: Iterable[MutationRecord[T]],
        crossover_records: Iterable[CrossoverRecord[T]],
    ) -> None:
        """
        Parameters:
            fitness_values:
                The records of molecules in the generation.
            mutation_records (list[MutationRecord[T]]):
                The records of mutations done during the generation.
            crossover_records (list[CrossoverRecord[T]]):
                The records of crossover operations done during the
                generation.
        """
        self._fitness_values = dict(fitness_values)
        self._mutation_records = tuple(mutation_records)
        self._crossover_records = tuple(crossover_records)

    def get_fitness_values(self) -> dict[T, FitnessValues]:
        """
        Get the fitness values of the generation.

        Returns:
            The fitness values.
        """
        return dict(self._fitness_values)

    def get_molecule_records(self) -> Iterator[T]:
        """
        Yield the molecule records in the generation.

        Yields:
            A molecule record.
        """
        yield from self._fitness_values

    def get_mutation_records(self) -> Iterator[MutationRecord[T]]:
        """
        Yield the mutation records in the generation.

        Yields:
            A mutation record.
        """
        yield from self._mutation_records

    def get_crossover_records(self) -> Iterator[CrossoverRecord[T]]:
        """
        Yield the crossover records in the generation.

        Yields:
            A crossover record.
        """
        yield from self._crossover_records



def test_database(case_data):
    """
    Test a database.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the database to test and the value to put
        into it.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_database(
        database=case_data.database,
        molecule=case_data.molecule,
        value=case_data.value,
    )


def _test_database(database, molecule, value):
    """
    Test a database.

    Parameters
    ----------
    database : class:`.ValueDatabase`
        The database to test.

    molecule : :class:`.Molecule`
        The molecule to test.

    value : :class:`object`
        The value to put into the database.

    Returns
    -------
    None : :class:`NoneType`

    """

    database.put(molecule, value)
    assert database.get(molecule) == value



import itertools as it


def test_get_core_atom_ids(case_data):
    """
    Test :meth:`.BuildingBlock.get_core_atom_ids`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the building block to test and the correct
        core atom ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    for id1, id2 in it.zip_longest(
        case_data.building_block.get_core_atom_ids(),
        case_data.core_atom_ids,
    ):
        assert id1 == id2



import itertools as it

from ..utilities import (
    are_clone_sequences,
    are_same_id_sequences,
    atom_id,
)


def test_get_functional_groups(case_data):
    """
    Test :meth:`.GenericFunctionalGroupFactory.get_functional_groups`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case. Holds the factory, molecule and correct
        functional groups.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_functional_groups(
        factory=case_data.factory,
        molecule=case_data.molecule,
        functional_groups=case_data.functional_groups,
    )


def _test_get_functional_groups(factory, molecule, functional_groups):
    """
    Test :meth:`.GenericFunctionalGroupFactory.get_functional_groups`.

    Parameters
    ----------
    factory : :class:`.GenericFunctionalGroupFactory`
        The factory to test.

    molecule : :class:`.Molecule`
        The molecule to test.

    functional_groups : :class:`tuple`
        The correct :class:`.GenericFunctionalGroup` instances.

    Returns
    -------
    None : :class:`NoneType`

    """

    for expected_fg, fg in it.zip_longest(
        functional_groups,
        factory.get_functional_groups(molecule),
    ):
        are_clone_functional_groups(expected_fg, fg)


def are_clone_functional_groups(functional_group1, functional_group2):
    """
    Test if the functional groups are clones of each other.

    """

    assert functional_group1.__class__ is functional_group2.__class__

    are_clone_sequences(
        atoms1=sorted(functional_group1.get_atoms(), key=atom_id),
        atoms2=sorted(functional_group2.get_atoms(), key=atom_id),
    )
    are_same_id_sequences(
        ids1=sorted(functional_group1.get_atom_ids()),
        ids2=sorted(functional_group2.get_atom_ids()),
    )

    are_same_id_sequences(
        ids1=sorted(functional_group1.get_placer_ids()),
        ids2=sorted(functional_group2.get_placer_ids()),
    )

    are_same_id_sequences(
        ids1=sorted(functional_group1.get_core_atom_ids()),
        ids2=sorted(functional_group2.get_core_atom_ids()),
    )

    are_clone_sequences(
        atoms1=sorted(functional_group1.get_bonders(), key=atom_id),
        atoms2=sorted(functional_group2.get_bonders(), key=atom_id),
    )
    are_same_id_sequences(
        ids1=sorted(functional_group1.get_bonder_ids()),
        ids2=sorted(functional_group2.get_bonder_ids()),
    )

    are_clone_sequences(
        atoms1=sorted(functional_group1.get_deleters(), key=atom_id),
        atoms2=sorted(functional_group2.get_deleters(), key=atom_id),
    )
    are_same_id_sequences(
        ids1=sorted(functional_group1.get_deleter_ids()),
        ids2=sorted(functional_group2.get_deleter_ids()),
    )



from functools import lru_cache

from stk._internal.key_makers.inchi_key import InchiKey

from ..value import ValueDatabase
from .utilities import HashableDict


class ValueMongoDb(ValueDatabase):
    """
    Use MongoDB to store and retrieve molecular property values.

    Examples
    --------
    See also examples in :class:`.ValueDatabase`.

    *Storing Molecular Properties in a Database*

    You want to store property values in a database.

    .. testsetup:: storing-molecular-properties-in-a-database

        import stk

        # Change the database used, so that when a developer
        # runs the doctests locally, their "stk" database is not
        # contaminated.
        _test_database = '_stk_doctest_database'
        _old_init = stk.ValueMongoDb
        stk.ValueMongoDb = lambda mongo_client, collection: (
            _old_init(
                mongo_client=mongo_client,
                database=_test_database,
                collection=collection,
            )
        )

        # Change the database MongoClient will connect to.

        import os
        import pymongo

        _mongo_client = pymongo.MongoClient
        _mongodb_uri = os.environ.get(
            'MONGODB_URI',
            'mongodb://localhost:27017/'
        )
        pymongo.MongoClient = lambda: _mongo_client(_mongodb_uri)

    .. testcode:: storing-molecular-properties-in-a-database

        import stk
        import pymongo

        # Connect to a MongoDB. This example connects to a local
        # MongoDB, but you can connect to a remote DB too with
        # MongoClient() - read the documentation for pymongo to see how
        # to do that.
        client = pymongo.MongoClient()
        db = stk.ValueMongoDb(
            mongo_client=client,
            collection='atom_counts',
        )

        molecule = stk.BuildingBlock('BrCCBr')
        # Add the value to the database.
        db.put(molecule, molecule.get_num_atoms())
        # Retrieve the value from the database.
        num_atoms = db.get(molecule)

        # Works with constructed molecules too.
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                ),
                repeating_unit='A',
                num_repeating_units=2,
            ),
        )
        db.put(polymer, polymer.get_num_atoms())
        num_polymer_atoms = db.get(polymer)

    .. testcode:: storing-molecular-properties-in-a-database
        :hide:

        assert num_polymer_atoms == polymer.get_num_atoms()

    .. testcleanup:: storing-molecular-properties-in-a-database

        stk.ValueMongoDb = _old_init
        pymongo.MongoClient().drop_database(_test_database)
        pymongo.MongoClient = _mongo_client

    """

    def __init__(
        self,
        mongo_client,
        collection,
        database="stk",
        key_makers=(InchiKey(),),
        put_lru_cache_size=128,
        get_lru_cache_size=128,
        indices=("InChIKey",),
    ):
        """
        Initialize a :class:`.ValueMongoDb` instance.

        Parameters
        ----------
        mongo_client : :class:`pymongo.MongoClient`
            The database client.

        collection : :class:`str`
            The name of the MongoDB collection used for storing the
            property values.

        database : :class:`str`, optional
            The name of the MongoDB database used for storing the
            property values.

        key_makers : :class:`tuple` of :class:`.MoleculeKeyMaker`
            Used to make the keys of molecules, which the values
            are associated with. If two molecules have the same
            key, they will return the same value from the database.

        put_lru_cache_size : :class:`int`, optional
            A RAM-based least recently used cache is used to avoid
            writing to the database repeatedly. This sets
            the number of values which fit into the LRU cache. If
            ``None``, the cache size will be unlimited.

        get_lru_cache_size : :class:`int`, optional
            A RAM-based least recently used cache is used to avoid
            reading from the database repeatedly. This sets
            the number of values which fit into the LRU cache. If
            ``None``, the cache size will be unlimited.

        indices : :class:`tuple` of :class:`str`, optional
            The names of molecule keys, on which an index should be
            created, in order to minimize lookup time.

        """

        self._values = mongo_client[database][collection]
        self._key_makers = key_makers
        self._put = lru_cache(maxsize=put_lru_cache_size)(self._put)
        self._get = lru_cache(maxsize=get_lru_cache_size)(self._get)

        index_information = self._values.index_information()
        if "v_1" not in index_information:
            self._values.create_index("v")

        for index in indices:
            # Do not create the same index twice.
            if f"{index}_1" not in index_information:
                self._values.create_index(index)

    def put(self, molecule, value):
        json = {"v": value}
        for key_maker in self._key_makers:
            json[key_maker.get_key_name()] = key_maker.get_key(molecule)
        # lru_cache requires that the parameters to the cached function
        # are hashable objects.
        return self._put(HashableDict(json))

    def _put(self, json):
        keys = dict(json)
        keys.pop("v")

        query = {"$or": []}
        for key, value in keys.items():
            query["$or"].append({key: value})

        self._values.update_many(
            filter=query,
            update={"$set": json},
            upsert=True,
        )

    def get(self, molecule):
        def make_dict(key_maker):
            return HashableDict(
                {key_maker.get_key_name(): key_maker.get_key(molecule)}
            )

        key = {"$or": tuple(map(make_dict, self._key_makers))}
        # lru_cache requires that the parameters to the cached function
        # are hashable objects.
        return self._get(HashableDict(key))

    def _get(self, key):
        value = self._values.find_one(key)
        if value is None:
            raise KeyError(
                "No molecule found in the database with a key of: " f"{key}"
            )
        return value["v"]



import pytest
import stk

from ....case_data import CaseData
from ...building_blocks import (
    get_other_linker,
    get_palladium_cispbi_sqpl,
)


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M3L3Triangle(
                    corners=get_palladium_cispbi_sqpl(),
                    linkers=get_other_linker(),
                    reaction_factory=stk.DativeReactionFactory(
                        reaction_factory=stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.GenericFunctionalGroup,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C2C([H])=C([H])N(->[Pd+2]3(<-N4=C([H])C([H])=C"
                "(C([H])=C4[H])C4=C([H])C([H])=N(->[Pd+2]5(<-N6=C([H]"
                ")C([H])=C(C([H])=C6[H])C6=C([H])C([H])=N(->[Pd+2]7(<"
                "-N8=C([H])C([H])=C2C([H])=C8[H])<-N([H])([H])C([H])("
                "[H])C([H])([H])N->7([H])[H])C([H])=C6[H])<-N([H])([H"
                "])C([H])([H])C([H])([H])N->5([H])[H])C([H])=C4[H])<-"
                "N([H])([H])C([H])([H])C([H])([H])N->3([H])[H])=C1[H]"
            ),
            name=name,
        ),
        # Non-metal-containing tests with increasing angles between
        # functional groups.
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M3L3Triangle(
                    corners=stk.BuildingBlock(
                        smiles="C1CC(C1Br)Br",
                        functional_groups=[stk.BromoFactory()],
                    ),
                    linkers=stk.BuildingBlock(
                        smiles="C(#CBr)Br",
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
            ),
            smiles=(
                "[H]C1([H])C([H])([H])C2([H])C#CC3([H])C([H])([H])C([H"
                "])([H])C3([H])C#CC3([H])C([H])([H])C([H])([H])C3([H])"
                "C#CC12[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M3L3Triangle(
                    corners=stk.BuildingBlock(
                        smiles="C1CC(CC(C1)Br)Br",
                        functional_groups=[stk.BromoFactory()],
                    ),
                    linkers=stk.BuildingBlock(
                        smiles="C(#CBr)Br",
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
            ),
            smiles=(
                "[H]C1([H])C([H])([H])C2([H])C#CC3([H])C([H])([H])C([H"
                "])([H])C([H])([H])C([H])(C#CC4([H])C([H])([H])C([H])("
                "[H])C([H])([H])C([H])(C#CC([H])(C1([H])[H])C2([H])[H]"
                ")C4([H])[H])C3([H])[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M3L3Triangle(
                    corners=stk.BuildingBlock(
                        smiles="C(CBr)Br",
                        functional_groups=[stk.BromoFactory()],
                    ),
                    linkers=stk.BuildingBlock(
                        smiles="C(#CBr)Br",
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
            ),
            smiles=(
                "[H]C1([H])C#CC([H])([H])C([H])([H])C#CC([H])([H])C([H"
                "])([H])C#CC1([H])[H]"
            ),
            name=name,
        ),
    ),
)
def metal_cage_m3l3_triangle(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )
